import io
import socket

import mock
import pytest

from ..guestagent import ClientError, GuestAgent


@pytest.fixture
def ga(monkeypatch):
    ga = GuestAgent("testvm", 0.1)
    ga.file = io.StringIO()
    ga.client = mock.MagicMock()
    randint = mock.Mock(return_value=87643)
    monkeypatch.setattr("random.randint", randint)
    return ga


def test_ga_read(ga):
    ga.file = io.StringIO('{"return": 17035}\n')
    assert 17035 == ga.read()


def test_ga_read_error(ga):
    ga.file = io.StringIO('{"return": 0, "error": "test failure"}\n')
    with pytest.raises(ClientError):
        ga.read()


def test_ga_sync_immediate(ga):
    ga.file = io.StringIO('{"return": 87643}\n')
    ga.sync()
    assert True


def test_ga_sync_retry(ga):
    ga.file = io.StringIO('{"return": 2}\n{"return": 87643}\n')
    ga.sync()
    assert True


def test_ga_sync_too_often(ga):
    ga.file = io.StringIO(
        """\
{"return": 2}
{"return": 3}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 4}
{"return": 87643}
"""
    )
    with pytest.raises(ClientError):
        ga.sync()


def test_ga_contextmgr(ga, monkeypatch, tmpdir):
    monkeypatch.setattr(socket, "socket", mock.MagicMock(socket.socket))
    with open(str(tmpdir / "socket"), "w") as f:
        f.write('{"return": 87643}\n')
    f = open(str(tmpdir / "socket"), "r")
    socket.socket().makefile.return_value = f
    with ga as g:
        assert g.machine == "testvm"



import configparser
import os.path


class SysConfig(object):
    """A global config state registry.

    This is used to manage system-specific configuration overrides
    for values used in various places within fc.qemu.

    This provides the code for loading all those options from a central
    config file and to allow tests overriding those values gracefully.
    """

    def __init__(self):
        self.qemu = {}
        self.ceph = {}
        self.agent = {}
        self.cp = None

    def read_config_files(self):
        """Tries to open fc-qemu.conf at various location."""
        self.cp = configparser.ConfigParser()
        self.cp.read(os.path.dirname(__file__) + "/default.conf")
        self.cp.read("/etc/qemu/fc-qemu.conf")
        if "qemu" not in self.cp.sections():
            raise RuntimeError(
                "error while reading config file: " "section [qemu] not found"
            )

    def load_system_config(self):
        self.read_config_files()

        self.qemu["migration_address"] = self.cp.get(
            "qemu", "migration-address"
        )
        self.qemu["require_kvm"] = bool(self.cp.get("qemu", "accelerator"))
        self.qemu["vnc"] = self.cp.get("qemu", "vnc")
        self.qemu["max_downtime"] = self.cp.getfloat("qemu", "max-downtime")
        self.qemu["migration_bandwidth"] = self.cp.getint(
            "qemu", "migration-bandwidth"
        )
        self.qemu["vm_max_total_memory"] = self.cp.getint(
            "qemu", "vm-max-total-memory"
        )
        self.qemu["vm_expected_overhead"] = self.cp.getint(
            "qemu", "vm-expected-overhead"
        )

        self.qemu["throttle_by_pool"] = tbp = {}
        for pool, iops in self.cp.items("qemu-throttle-by-pool"):
            tbp[pool] = int(iops)

        # Consul
        self.agent["consul_token"] = self.cp.get("consul", "access-token")
        self.agent["consul_event_threads"] = self.cp.getint(
            "consul", "event-threads"
        )

        # Qemu
        self.agent["accelerator"] = self.cp.get("qemu", "accelerator")
        self.agent["machine_type"] = self.cp.get("qemu", "machine-type")
        self.agent["migration_ctl_address"] = self.cp.get(
            "qemu", "migration-ctl-address"
        )
        self.agent["binary_generation"] = self.cp.getint(
            "qemu", "binary-generation"
        )
        self.agent["timeout_graceful"] = self.cp.getint(
            "qemu", "timeout-graceful"
        )
        self.agent["vhost"] = self.cp.getboolean("qemu", "vhost")

        # Ceph
        self.agent["this_host"] = self.cp.get("ceph", "lock_host")
        self.agent["ceph_id"] = self.cp.get("ceph", "client-id")

        self.ceph["CEPH_CLIENT"] = self.cp.get(
            "ceph", "client-id", fallback="admin"
        )
        self.ceph["CEPH_CLUSTER"] = self.cp.get(
            "ceph", "cluster", fallback="ceph"
        )
        self.ceph["CEPH_CONF"] = self.cp.get("ceph", "ceph-conf")
        self.ceph["CEPH_LOCK_HOST"] = self.cp.get("ceph", "lock_host")
        self.ceph["CREATE_VM"] = self.cp.get("ceph", "create-vm")
        self.ceph["MKFS_XFS"] = self.cp.get("ceph", "mkfs-xfs")
        self.ceph["MKFS_EXT4"] = self.cp.get("ceph", "mkfs-ext4")


sysconfig = SysConfig()



import json
import os.path
import urllib.parse
import xmlrpc.client


def load_default_enc_json():
    if os.path.exists("/etc/nixos/enc.json"):
        with open("/etc/nixos/enc.json") as f:
            return json.load(f)
    else:
        with open("/etc/puppet/enc.json") as f:
            data = json.load(f)
        with open("/etc/directory.secret") as f:
            data["parameters"]["directory_password"] = f.read().strip()
        return data
    raise RuntimeError("No ENC file found.")


def connect(enc=None, ring=1):
    """Returns XML-RPC directory connection.

    The directory secret is read from `/etc/nixos/enc.json`.
    Alternatively, the parsed JSON content can be passed directly as
    dict.

    Selects ring0/ring1 API according to the `ring` parameter. Giving `max`
    results in selecting the highest ring available according to the ENC.
    """
    if not enc:
        enc = load_default_enc_json()
    if ring == "max":
        ring = enc["parameters"]["directory_ring"]
    base_url = enc["parameters"].get(
        "directory_url", "https://directory.fcio.net/v2/api"
    )
    url_parts = urllib.parse.urlsplit(base_url)

    url = (
        # fmt: off
        url_parts.scheme + "://"
        + enc["name"] + ":" + enc["parameters"]["directory_password"] + "@"
        + url_parts.netloc + url_parts.path
        # fmt: on
    )
    if ring == 1:
        url += "/rg-" + enc["parameters"]["resource_group"]

    return xmlrpc.client.ServerProxy(url, allow_none=True, use_datetime=True)



import mock
import pytest

from fc.qemu.exc import MigrationError
from fc.qemu.incoming import (
    IncomingAPI,
    IncomingServer,
    authenticated,
    parse_address,
)


def test_authentication_wrapper():
    @authenticated
    def test(cookie):
        return 1

    context = mock.Mock()
    context.cookie = "asdf"
    assert test(context, "asdf") == 1

    with pytest.raises(MigrationError):
        test(context, "foobar")


def test_parse_address_host():
    assert ("host", 1234) == parse_address("host:1234")


def test_parse_address_ipv4():
    assert ("10.1.2.3", 4567) == parse_address("10.1.2.3:4567")


def test_parse_address_ipv5():
    assert ("2001:db8:1::45", 1234) == parse_address("[2001:db8:1::45]:1234")


def test_incoming_api():
    server = mock.Mock()
    server.agent.ceph.auth_cookie.return_value = "asdf"
    api = IncomingAPI(server)
    assert api.cookie == "asdf"

    api.ping("asdf")
    assert server.extend_cutoff_time.call_args_list == [
        mock.call(hard_timeout=60)
    ]

    api.acquire_migration_lock("asdf")
    assert server.acquire_migration_lock.call_args_list == [mock.call()]

    api.release_migration_lock("asdf")
    assert server.release_migration_lock.call_args_list == [mock.call()]

    api.acquire_ceph_locks("asdf")
    assert server.acquire_ceph_locks.call_args_list == [mock.call()]

    api.prepare_incoming("asdf", [], {})
    assert server.prepare_incoming.call_args_list == [mock.call([], {})]

    api.finish_incoming("asdf")
    assert server.finish_incoming.call_args_list == [mock.call()]

    api.rescue("asdf")
    assert server.rescue.call_args_list == [mock.call()]

    api.destroy("asdf")
    assert server.destroy.call_args_list == [mock.call()]

    api.cancel("asdf")
    assert server.cancel.call_args_list == [mock.call()]


def test_incoming_server():
    agent = mock.Mock()
    agent.migration_ctl_address = "localhost:9000"
    server = IncomingServer(agent)
    assert server.bind_address == ("localhost", 9000)

    server._now = server.timeout._now = mock.Mock(return_value=30)
    server.timeout.cutoff = 29
    server.extend_cutoff_time(soft_timeout=60)
    assert server.timeout.cutoff == 90
    server.extend_cutoff_time(soft_timeout=30)
    assert server.timeout.cutoff == 90
    server.extend_cutoff_time(soft_timeout=40)
    assert server.timeout.cutoff == 90
    server.extend_cutoff_time(soft_timeout=120)
    assert server.timeout.cutoff == 150
    server.extend_cutoff_time(hard_timeout=30)
    assert server.timeout.cutoff == 60
    server.extend_cutoff_time(hard_timeout=340)
    assert server.timeout.cutoff == 370
    server.extend_cutoff_time(soft_timeout=30)
    assert server.timeout.cutoff == 370



import pprint
import random
import threading
import time
import xmlrpc.client

from .exc import ConfigChanged
from .timeout import TimeOut


class Heartbeat(object):
    """Continuously ping a target in the background to indicate that
    we're still here.

    Run this in a separate thread so that blocks from other things (like
    the Qemu monitor) do not cause the remote side to time out and thus
    block any chance for recovery.

    """

    cookie = None

    def __init__(self, log):
        self.thread = threading.Thread(target=self.run)
        self.running = False
        self.url = None
        self.log = log
        self.log.debug("heartbeat-initialized")

    def start(self):
        assert self.url is not None
        if self.running:
            return
        self.connection = xmlrpc.client.ServerProxy(self.url, allow_none=True)
        self.running = True
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        self.running = False

    def run(self):
        self.log.debug("started-heartbeat-ping")
        try:
            while self.running:
                try:
                    self.log.debug("heartbeat-ping")
                    self.connection.ping(self.cookie)
                except Exception:
                    self.log.exception("ping-failed", exc_info=True)
                time.sleep(10)
        finally:
            self.log.debug("stopped-heartbeat-ping")


class Outgoing(object):

    migration_exitcode = None
    target = None
    cookie = None

    # How long to wait until we discover an inmigrate service?
    # This should happen relatively fast, but just in case we'll keep a high
    # timeout. The agents should be spawned relatively quickly everywhere.
    connect_timeout = 60 * 60  # 1 hour

    # How long to wait until we get a migration lock?
    # This can happen quite slowly as during busy times migration may be slow
    # and some large and busy VMs can take 1-2hours (on Gigabit). If this adds
    # up over multiple hosts I'm giving a grace period of up to 12 hours here.
    migration_lock_timeout = 12 * 60 * 60  # 12 hours.

    def __init__(self, agent):
        self.agent = agent
        self.log = agent.log
        self.name = agent.name
        self.consul = agent.consul
        self.heartbeat = Heartbeat(self.log)

    def __call__(self):
        self.cookie = self.agent.ceph.auth_cookie()
        with self:
            self.connect()
            self.acquire_migration_locks()
            self.transfer_ceph_locks()
            self.migrate()

        return self.migration_exitcode

    # The context manager is suitable to express "I am going to
    # interact with this VM: please prepare connections to
    # the according backends and make sure to clean up after me.""
    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        if _exc_type is None:
            # We're on our happy path. Release the migration lock
            # optimistically and only in this case. Otherwise the
            # error reporting for why the migration has failed will
            # be screwed. And the negative path will exit the process
            # and eventually give up locks anyway.
            try:
                self.agent.qemu.release_migration_lock()
            except Exception:
                pass
            try:
                self.target.release_migration_lock(self.cookie)
            except Exception:
                pass

            self.migration_exitcode = 0

        else:
            self.migration_exitcode = 1
            try:
                # XXX silence timeout errors here to avoid unnecessary
                # traceback output
                self.log.exception(
                    "migration-failed", action="rescue", exc_info=True
                )
                self.rescue()
                return True  # swallow exception
            except:  # noqa
                # Purposeful bare except: try really hard to kill
                # our VM.
                self.log.exception(
                    "rescue-failed", exc_info=True, action="destroy"
                )
                self.destroy()

    def locate_inmigrate_service(self):
        service_name = "vm-inmigrate-" + self.name
        timeout = TimeOut(
            self.connect_timeout,
            interval=3,
            raise_on_timeout=True,
            log=self.log,
        )
        self.log.info("locate-inmigration-service")
        while timeout.tick():
            if self.agent.has_new_config():
                raise ConfigChanged()
            candidates = self.consul.catalog.service(service_name)
            if len(candidates) > 1:
                self.log.warning(
                    "multiple-services-found",
                    action="trying newest first",
                    service=service_name,
                )
            candidates = sorted(
                candidates, key=lambda i: i["ModifyIndex"], reverse=True
            )
            for candidate in candidates:
                url = "http://{}:{}".format(
                    candidate["ServiceAddress"], candidate["ServicePort"]
                )
                self.log.info("located-inmigration-service", url=url)
                try:
                    target = xmlrpc.client.ServerProxy(url, allow_none=True)
                    target.ping(self.cookie)
                except Exception as e:
                    self.log.info("connect", result="failed", reason=str(e))
                    # Not a viable service. Delete it.
                    # XXX This seems broken in consulate 0.6, waiting
                    # for the 1.0 release ...
                    # try:
                    #     self.consul.catalog.deregister(
                    #        node=candidate['Node'],
                    #        datacenter=candidate['Datacenter'],
                    #        service_id=candidate['ServiceID'])
                    #     self.log.debug('delete-stale-incoming-service', result='success')
                    #     pass
                    # except Exception:
                    #     self.log.exception('delete-stale-incoming-service', result='failed', exc_info=True)
                    pass
                else:
                    break
            else:
                # We did not find a viable target - continue waiting
                continue
            return target, url
        raise RuntimeError("failed to locate inmigrate service", service_name)

    def connect(self):
        connection, url = self.locate_inmigrate_service()
        self.target = connection
        self.heartbeat.url = url
        self.heartbeat.cookie = self.cookie
        self.heartbeat.start()

    def acquire_migration_locks(self):
        tries = 0
        self.log.info("acquire-migration-locks")
        timeout = TimeOut(
            self.migration_lock_timeout,
            interval=3,
            raise_on_timeout=True,
            log=self.log,
        )
        while timeout.tick():
            # In case that there are multiple processes waiting, randomize to
            # avoid steplock retries. We us CSMA/CD-based exponential backoff,
            # timeslot 10ms but with a max of 13 instead of 16.
            # This means we'll wait up to 80s, 40s on average if everything
            # becomes really busy and we may experience lock contention.
            tries = min([tries + 1, 13])
            timeout.interval = random.randint(1, 2**tries) * 0.01
            if self.agent.has_new_config():
                self.target.cancel(self.cookie)
                raise ConfigChanged()

            # Try to acquire local lock
            if self.agent.qemu.acquire_migration_lock():
                self.log.debug("acquire-local-migration-lock", result="success")
            else:
                self.log.debug("acquire-local-migration-lock", result="failure")
                continue

            # Try to acquire remote lock
            try:
                self.log.debug("acquire-remote-migration-lock")
                # We got our lock, now ask the remote side:
                if not self.target.acquire_migration_lock(self.cookie):
                    self.log.debug(
                        "acquire-remote-migration-lock", result="failure"
                    )
                    self.agent.qemu.release_migration_lock()
                    continue
                self.log.debug(
                    "acquire-remote-migration-lock", result="success"
                )
            except Exception:
                self.log.exception(
                    "acquire-remote-migration-lock",
                    result="failure",
                    exc_info=True,
                )
                self.agent.qemu.release_migration_lock()
                continue
            break

    def transfer_ceph_locks(self):
        self.agent.ceph.unlock()
        self.target.acquire_ceph_locks(self.cookie)

    def migrate(self):
        """Actually move VM between hosts."""
        args, config = self.agent.qemu.get_running_config()
        self.log.info("prepare-remote-environment")
        migration_address = self.target.prepare_incoming(
            self.cookie, args, config
        )
        self.log.info("start-migration", target=migration_address)
        self.agent.qemu.migrate(migration_address)
        try:
            for stat in self.agent.qemu.poll_migration_status():
                remaining = stat["ram"]["remaining"] if "ram" in stat else 0
                mbps = stat["ram"]["mbps"] if "ram" in stat else "-"
                self.log.info(
                    "migration-status",
                    status=stat["status"],
                    remaining="{0:,d}".format(remaining),
                    mbps=mbps,
                    output=pprint.pformat(stat),
                )
        except Exception:
            self.log.exception("error-waiting-for-migration", exc_info=True)
            raise

        status = self.agent.qemu.qmp.command("query-status")
        assert not status["running"], status
        assert status["status"] == "postmigrate", status
        self.log.info("finish-migration")
        try:
            self.heartbeat.stop()
            self.target.finish_incoming(self.cookie)
        except Exception:
            self.log.exception("error-finish-remote", exc_info=True)
        self.agent.qemu.destroy()

    def rescue(self):
        """Outgoing rescue: try to rescue the remote side first."""
        if self.target is not None:
            try:
                self.log.info("rescue-remote")
                self.target.rescue(self.cookie)
                self.target.finish_incoming(self.cookie)
                self.log.info("rescue-remote-success", action="destroy local")
                self.destroy()
                # We managed to rescue on the remote side - hooray!
                self.migration_exitcode = 0
                return
            except Exception:
                self.log.exception(
                    "rescue-remote-failed",
                    exc_info=True,
                    action="destroy remote",
                )
                try:
                    self.heartbeat.stop()
                    self.target.destroy(self.cookie)
                except Exception:
                    self.log.exception("destroy-remote-failed", exc_info=True)
        try:
            self.log.info("continue-locally")
            self.agent.ceph.lock()
            assert self.agent.qemu.is_running()
        except Exception:
            self.log.exception(
                "continue-locally",
                exc_info=True,
                result="failed",
                action="destroy local",
            )
            self.destroy()
        else:
            self.log.info("continue-locally", result="success")

    def destroy(self):
        self.agent.qemu.destroy()
        self.agent.qemu.clean_run_files()



import pytest
import rbd


@pytest.fixture
def ceph_with_volumes(ceph_inst):
    for vol in ceph_inst.volumes:
        vol.ensure_presence()
    ceph_inst.lock()
    yield ceph_inst
    for vol in ceph_inst.volumes:
        vol.unlock(force=True)
        vol.snapshots.purge()
        if vol._image:
            vol._image.close()
        rbd.RBD().remove(ceph_inst.ioctx, vol.name)


def test_ceph_stop_should_unlock_all_volumes(ceph_with_volumes):
    for vol in ceph_with_volumes.volumes:
        assert vol.lock_status()
    ceph_with_volumes.stop()
    for vol in ceph_with_volumes.volumes:
        assert vol.lock_status() is None


def test_ceph_stop_remove_only_own_locks(ceph_with_volumes):
    """Test case where failed migrations leave inconsistent locking."""
    ceph_with_volumes.root.unlock()
    ceph_with_volumes.root.rbdimage.lock_exclusive("someotherhost")
    # It unlocks what it can.
    ceph_with_volumes.stop()
    assert ceph_with_volumes.root.lock_status()
    assert ceph_with_volumes.swap.lock_status() is None
    assert ceph_with_volumes.tmp.lock_status() is None


def test_is_unlocked(ceph_with_volumes):
    assert ceph_with_volumes.is_unlocked() is False
    ceph_with_volumes.unlock()
    assert ceph_with_volumes.is_unlocked() is True



import mock
import pytest

from ..exc import MigrationError
from ..incoming import IncomingAPI, IncomingServer


@pytest.fixture
def mock_agent():
    agent = mock.Mock()
    agent.migration_ctl_address = "localhost:12345"
    agent.qemu = mock.Mock()
    agent.ceph = mock.Mock()
    agent.ceph.auth_cookie.return_value = "5f620fda"
    return agent


def test_prepare_should_stop_ceph_on_exception(mock_agent):
    mock_agent.qemu.inmigrate.side_effect = Exception("boom!")
    s = IncomingServer(mock_agent)
    with pytest.raises(Exception):
        s.prepare_incoming("args", "config")
    assert mock_agent.ceph.stop.called is True


def test_rescue(mock_agent):
    s = IncomingServer(mock_agent)
    mock_agent.ceph.lock.side_effect = Exception("boom!")
    with pytest.raises(Exception):
        s.rescue()
    assert mock_agent.qemu.destroy.called is True
    assert mock_agent.ceph.unlock.called is True


@mock.patch("fc.qemu.incoming.IncomingServer")
def test_authentication_match(server):
    api = IncomingAPI(server)
    api.cookie = "cookie1"
    # should not raise an exception
    assert api.ping("cookie1") is None


@mock.patch("fc.qemu.incoming.IncomingServer")
def test_authentication_mismatch(server):
    api = IncomingAPI(server)
    api.cookie = "cookie1"
    with pytest.raises(MigrationError):
        assert api.ping("cookie-does-not-match") is None


def test_screen_config_disable_iommu(mock_agent):
    s = IncomingServer(mock_agent)
    assert (
        s.screen_config(
            """\
[machine]
  type = "pc-i440fx-2.5"
  iommu = "off"
  accel = "kvm"

"""
        )
        == """\
[machine]
  type = "pc-i440fx-2.5"

  accel = "kvm"

"""
    )


def test_screen_config_update_qmp_monitor_syntax(mock_agent):
    s = IncomingServer(mock_agent)
    assert (
        s.screen_config(
            """\

# QMP monitor support via Unix socket

[mon "qmp_monitor"]
  mode = "control"
  chardev = "ch_qmp_monitor"
  default = "on"

[chardev "ch_qmp_monitor"]
  backend = "socket"
  path = "/run/qemu.{name}.qmp.sock"
  server = "on"
  wait = "off"

"""
        )
        == """\

# QMP monitor support via Unix socket

[mon "qmp_monitor"]
  mode = "control"
  chardev = "ch_qmp_monitor"
  pretty = "off"

[chardev "ch_qmp_monitor"]
  backend = "socket"
  path = "/run/qemu.{name}.qmp.sock"
  server = "on"
  wait = "off"

"""
    )



class LockError(RuntimeError):
    pass


class MigrationError(RuntimeError):
    pass


class DestructionError(RuntimeError):
    """Failed to destroy a running VM."""

    pass


class QemuNotRunning(Exception):
    """Something happened and we're sure Qemu isn't running."""

    pass


class InvalidCommand(RuntimeError):
    pass


class VMConfigNotFound(RuntimeError):
    pass


class VMStateInconsistent(RuntimeError):

    qemu = None
    proc = None
    ceph_lock = None

    def is_consistent(self):
        states = [self.qemu, self.proc, self.ceph_lock]
        return any(states) == all(states)


class ConfigChanged(Exception):
    """Used to break up flows to signal that the config changed and we
    may want to retry.

    Useful to break up long waiting periods while data is changing.

    """



import os
import subprocess
import time

import psutil
import pytest

from ..hazmat.qemu import Qemu


@pytest.fixture
def qemu_with_pidfile(tmpdir):
    # The fixture uses a very long name which catches a special case where
    # the proc name is limited to 16 bytes and then we failed to match
    # the running VM. Parametrizing fixtures doesn't work the way I want
    # so I did it this way ...
    pidfile = str(tmpdir / "run/qemu.testvmwithverylongname.pid")
    try:
        os.unlink(pidfile)
    except OSError:
        pass
    proc = subprocess.Popen(
        [
            "qemu-system-x86_64",
            "-name",
            "testvmwithverylongname,process=kvm.testvmwithverylongname",
            "-nodefaults",
            "-pidfile",
            pidfile,
        ]
    )
    while not os.path.exists(pidfile):
        time.sleep(0.01)
    q = Qemu(dict(name="testvmwithverylongname", id=1234))
    try:
        yield q
    finally:
        proc.kill()


def test_proc_running(qemu_with_pidfile):
    assert isinstance(qemu_with_pidfile.proc(), psutil.Process)


def test_proc_not_running(qemu_with_pidfile):
    with open(qemu_with_pidfile.pidfile, "w") as p:
        p.write("0\n")
    assert qemu_with_pidfile.proc() is None


def test_proc_wrong_process(qemu_with_pidfile):
    with open(qemu_with_pidfile.pidfile, "w") as p:
        p.write("1\n")
    assert qemu_with_pidfile.proc() is None


def test_proc_no_pidfile(qemu_with_pidfile):
    os.unlink(qemu_with_pidfile.pidfile)
    assert qemu_with_pidfile.proc() is None


def test_proc_empty_pidfile(qemu_with_pidfile):
    open(qemu_with_pidfile.pidfile, "w").close()
    assert qemu_with_pidfile.proc() is None


def test_proc_pidfile_with_garbage(qemu_with_pidfile):
    """pidfiles are allowed to contain trailing lines with garbage,
    process retrieval must still work then."""
    with open(qemu_with_pidfile.pidfile, "a") as f:
        f.write("trailing line\n")
    assert isinstance(qemu_with_pidfile.proc(), psutil.Process)


def test_disk_cache_mode_default_writeback():
    q = Qemu(dict(name="testvm", id=1234))
    assert q.disk_cache_mode == "writeback"


def test_disk_cache_mode_default_writeback2():
    q = Qemu(dict(name="testvm", id=1234, qemu=dict()))
    assert q.disk_cache_mode == "writeback"


def test_disk_cache_mode_enc_enabled():
    q = Qemu(dict(name="testvm", id=1234, qemu=dict(write_back_cache=True)))
    assert q.disk_cache_mode == "writeback"


def test_disk_cache_mode_enc_disabled():
    q = Qemu(dict(name="testvm", id=1234, qemu=dict(write_back_cache=False)))
    assert q.disk_cache_mode == "none"



import json
import os.path
import typing
from codecs import encode
from io import StringIO

import pytest

from fc.qemu import util
from fc.qemu.agent import Agent
from fc.qemu.hazmat.qemu import Qemu

from ..conftest import get_log
from ..ellipsis import Ellipsis


@pytest.fixture(autouse=True)
def consul_timeouts(monkeypatch):
    monkeypatch.setattr(Qemu, "guestagent_timeout", 0.01)
    monkeypatch.setattr(Qemu, "qmp_timeout", 0.01)
    monkeypatch.setattr(Qemu, "thaw_retry_timeout", 0.1)


def test_no_events():
    stdin = StringIO("[]")
    Agent.handle_consul_event(stdin)
    assert util.log_data == []


def test_empty_event():
    stdin = StringIO(
        '[{"ModifyIndex": 123, "Value": "", "Key": "node/test22"}]'
    )
    Agent.handle_consul_event(stdin)
    assert util.log_data == [
        "start-consul-events count=1",
        "handle-key key=node/test22",
        "ignore-key key=node/test22 reason=empty value",
        "finish-consul-events",
    ]


def test_no_key_event():
    stdin = StringIO('[{"ModifyIndex": 123, "Value": ""}]')
    Agent.handle_consul_event(stdin)
    assert util.log_data == [
        "start-consul-events count=1",
        "handle-key-failed exc_info=True key=None",
        "finish-handle-key key=None",
        "finish-consul-events",
    ]


@pytest.fixture
def clean_config_test22(tmpdir):
    targets = [
        str(tmpdir / "etc/qemu/vm/test22.cfg"),
        str(tmpdir / "/etc/qemu/vm/.test22.cfg.staging"),
    ]
    for target in targets:
        if os.path.exists(target):
            os.unlink(target)
    yield
    for target in targets:
        if os.path.exists(target):
            os.unlink(target)


def prepare_consul_event(
    key: str = None, value: dict = None, index: int = 128, events: list = None
) -> typing.TextIO:
    """Generate a single event record (by passing key/str/index) or a list
    of event records (by passing events=[(key, value, index), ...])
    """
    if key is not None:
        events = [(key, value, index)]
    event_values = []
    for key, value, index in events:
        # Turn our payload into json and ensure a 7-bit ascii armour.
        # I'm not really sure why we introduced this in the first place, but I'm
        # keeping it this way to avoid inter-version compatibility issues.
        value = json.dumps(value).encode("ascii")
        value = encode(value, "base64")
        value = value.replace(b"\n", b"")
        # The ASCII armour needs to be turned into text again, because the JSON
        # encoder doesn't handle bytes-like objects.
        value = value.decode("ascii")
        event_values.append({"ModifyIndex": index, "Value": value, "Key": key})
    output = StringIO()
    json.dump(event_values, output)
    output.seek(0)
    return output


def test_qemu_config_change(clean_config_test22):
    util.test_log_options["show_methods"] = ["info"]
    cfg = {
        "classes": [
            "role::appserver",
            "role::backupclient",
            "role::generic",
            "role::postgresql90",
            "role::pspdf",
            "role::webproxy",
        ],
        "consul-generation": 128,
        "name": "test22",
        "parameters": {
            "ceph_id": "admin",
            "cores": 1,
            "directory_password": "1jmGYd3dddyjpFlLqg63RHie",
            "directory_ring": 1,
            "disk": 15,
            "environment": "staging",
            "id": 4097,
            "interfaces": {},
            "kvm_host": "host1",
            "location": "whq",
            "machine": "virtual",
            "memory": 512,
            "name": "test22",
            "online": False,
            "production": False,
            "profile": "generic",
            "rbd_pool": "rbd.ssd",
            "resource_group": "test",
            "resource_group_parent": "",
            "reverses": {},
            "service_description": "asdf",
            "servicing": True,
            "swap_size": 1073741824,
            "timezone": "Europe/Berlin",
            "tmp_size": 5368709120,
        },
    }

    first_event = prepare_consul_event("node/test22", cfg)
    Agent.handle_consul_event(first_event)

    assert util.log_data == [
        "start-consul-events count=1",
        "processing-consul-event consul_event=node machine=test22",
        "launch-ensure cmd=['true', '-D', 'ensure', 'test22'] machine=test22",
        "finish-consul-events",
    ]

    # The test doesn't really active the config, so we need to mock this.
    util.log_data = []
    agent = Agent("test22")
    assert agent.has_new_config()
    assert not os.path.exists(agent.configfile)
    agent.activate_new_config()
    agent._update_from_enc()
    assert not agent.has_new_config()
    assert os.path.exists(agent.configfile)
    assert util.log_data == []

    # Applying the same config again doesn't cause another ensure run.
    first_event.seek(0)
    util.log_data = []
    Agent.handle_consul_event(first_event)

    assert util.log_data == [
        "start-consul-events count=1",
        "processing-consul-event consul_event=node machine=test22",
        "ignore-consul-event machine=test22 reason=config is unchanged",
        "finish-consul-events",
    ]

    # Changing the config does cause ensure to be called.
    util.log_data = []
    cfg["disk"] = 20
    second_event = prepare_consul_event("node/test22", cfg, 135)
    Agent.handle_consul_event(second_event)
    assert util.log_data == [
        "start-consul-events count=1",
        "processing-consul-event consul_event=node machine=test22",
        "launch-ensure cmd=['true', '-D', 'ensure', 'test22'] machine=test22",
        "finish-consul-events",
    ]


def test_qemu_config_change_physical():
    util.test_log_options["show_events"] = ["consul"]

    cfg = {
        "classes": [
            "role::appserver",
            "role::backupclient",
            "role::generic",
            "role::postgresql90",
            "role::pspdf",
            "role::webproxy",
        ],
        "name": "test22",
        "parameters": {
            "ceph_id": "admin",
            "cores": 1,
            "directory_password": "1jmGYd3dddyjpFlLqg63RHie",
            "directory_ring": 1,
            "disk": 15,
            "environment": "staging",
            "id": 4097,
            "interfaces": {},
            "kvm_host": "host1",
            "location": "whq",
            "machine": "physical",
            "memory": 512,
            "name": "test22",
            "online": False,
            "production": False,
            "profile": "generic",
            "rbd_pool": "rbd.ssd",
            "resource_group": "test",
            "resource_group_parent": "",
            "reverses": {},
            "service_description": "asdf",
            "servicing": True,
            "swap_size": 1073741824,
            "timezone": "Europe/Berlin",
            "tmp_size": 5368709120,
        },
    }

    event = prepare_consul_event("node/test22", cfg, 123)
    Agent.handle_consul_event(event)
    assert util.log_data == [
        "start-consul-events count=1",
        "ignore-consul-event machine=test22 reason=is a physical machine",
        "finish-consul-events",
    ]


@pytest.mark.timeout(60)
@pytest.mark.live
def test_snapshot_online_vm(vm):
    util.test_log_options["show_events"] = [
        "consul",
        "snapshot",
        "thaw",
        "freeze",
        "retry",
        "fail",
    ]

    vm.ensure_online_local()
    vm.qemu.qmp.close()
    get_log()

    snapshot = {"vm": "simplevm", "snapshot": "backy-1234"}
    event = prepare_consul_event("snapshot/7468743", snapshot, 123)
    Agent.handle_consul_event(event)
    assert (
        Ellipsis(
            """\
start-consul-events count=1
snapshot machine=simplevm snapshot=backy-1234
snapshot-create machine=simplevm name=backy-1234
freeze machine=simplevm volume=root
freeze-failed action=continue machine=simplevm reason=Unable to sync with guest agent after 10 tries.
snapshot-ignore machine=simplevm reason=not frozen
ensure-thawed machine=simplevm volume=root
guest-fsfreeze-thaw-failed exc_info=True machine=simplevm subsystem=qemu
ensure-thawed-failed machine=simplevm reason=Unable to sync with guest agent after 10 tries.
handle-key-failed exc_info=True key=snapshot/7468743
finish-consul-events"""
        )
        == get_log()
    )


def test_snapshot_nonexisting_vm():
    util.test_log_options["show_events"] = ["consul", "unknown", "snapshot"]

    get_log()

    snapshot = {"vm": "test77", "snapshot": "backy-1234"}
    event = prepare_consul_event("snapshot/7468743", snapshot, 123)
    Agent.handle_consul_event(event)
    assert (
        get_log()
        == """\
start-consul-events count=1
unknown-vm machine=test77
snapshot-ignore machine=test77 reason=failed loading config snapshot=backy-1234
finish-consul-events"""
    )


@pytest.mark.live
@pytest.mark.timeout(60)
def test_snapshot_offline_vm(vm):
    util.test_log_options["show_events"] = ["consul", "snapshot"]

    vm.enc["parameters"]["kvm_host"] = "foobar"
    vm.stage_new_config()
    vm.activate_new_config()

    vm.ceph.ensure_root_volume()
    vm.ensure_offline()
    get_log()

    snapshot = {"vm": "simplevm", "snapshot": "backy-1234"}
    event = prepare_consul_event("snapshot/7468743", snapshot, 123)
    Agent.handle_consul_event(event)
    assert (
        get_log()
        == """\
start-consul-events count=1
snapshot machine=simplevm snapshot=backy-1234
snapshot expected=VM running machine=simplevm
finish-consul-events"""
    )


def test_multiple_events():
    util.test_log_options["show_events"] = ["consul", "handle-key"]

    cfg = {
        "classes": [
            "role::appserver",
            "role::backupclient",
            "role::generic",
            "role::postgresql90",
            "role::pspdf",
            "role::webproxy",
        ],
        "name": "test22",
        "parameters": {
            "ceph_id": "admin",
            "cores": 1,
            "directory_password": "1jmGYd3dddyjpFlLqg63RHie",
            "directory_ring": 1,
            "disk": 15,
            "environment": "staging",
            "id": 4097,
            "interfaces": {},
            "kvm_host": "host1",
            "location": "whq",
            "machine": "physical",
            "memory": 512,
            "name": "test22",
            "online": False,
            "production": False,
            "profile": "generic",
            "rbd_pool": "rbd.ssd",
            "resource_group": "test",
            "resource_group_parent": "",
            "reverses": {},
            "service_description": "asdf",
            "servicing": True,
            "swap_size": 1073741824,
            "timezone": "Europe/Berlin",
            "tmp_size": 5368709120,
        },
    }

    events = prepare_consul_event(events=[("node/test22", cfg, 123)] * 4)
    Agent.handle_consul_event(events)
    assert util.log_data == [
        "start-consul-events count=4",
        "handle-key key=node/test22",
        "ignore-consul-event machine=test22 reason=is a physical machine",
        "finish-handle-key key=node/test22",
        "handle-key key=node/test22",
        "ignore-consul-event machine=test22 reason=is a physical machine",
        "finish-handle-key key=node/test22",
        "handle-key key=node/test22",
        "ignore-consul-event machine=test22 reason=is a physical machine",
        "finish-handle-key key=node/test22",
        "handle-key key=node/test22",
        "ignore-consul-event machine=test22 reason=is a physical machine",
        "finish-handle-key key=node/test22",
        "finish-consul-events",
    ]



import errno
import os

import pytest
import rados
import rbd

from . import volume
from .ceph import Ceph


class RadosMock(object):
    def __init__(self, conffile, name):
        self.conffile = conffile
        self.name = name
        self._ioctx = {}
        self.__connected__ = False

    def connect(self):
        assert not self.__connected__
        self.__connected__ = True

    def open_ioctx(self, pool):
        if pool not in self._ioctx:
            self._ioctx[pool] = IoctxMock(pool)
        return self._ioctx[pool]

    def shutdown(self):
        assert self.__connected__
        self.__connected__ = False


class IoctxMock(object):
    """Mock access to a pool."""

    def __init__(self, name: str):
        # the rados implementation takes the name as a str, but later returns
        # that attribute as bytes
        self.name = name.encode("ascii")
        self.rbd_images = {}
        self._snapids = 0

    def _rbd_create(self, name, size):
        assert name not in self.rbd_images
        self.rbd_images[name] = dict(size=size, lock=None)

    def _rbd_create_snap(self, name, snapname):
        self._snapids += 1
        fullname = name + "@" + snapname
        assert fullname not in self.rbd_images
        self.rbd_images[fullname] = snap = self.rbd_images[name].copy()
        snap["lock"] = None
        snap["snapid"] = self._snapids

    def _rbd_remove(self, name):
        # XXX prohibit while snapshots exist and if locked/opened
        if name in self.rbd_images:
            del self.rbd_images[name]

    def _rbd_remove_snap(self, name, snapname):
        fullname = name + "@" + snapname
        if fullname in self.rbd_images:
            del self.rbd_images[fullname]

    def close(self):
        pass


class RBDMock(object):
    def list(self, ioctx):
        return list(ioctx.rbd_images.keys())

    def create(self, ioctx, name, size):
        ioctx._rbd_create(name, size)

    def remove(self, ioctx, name):
        ioctx._rbd_remove(name)


class ImageMock(object):
    def __init__(self, ioctx, name, snapname=None):
        self.ioctx = ioctx
        self.name = name
        self.snapname = snapname
        self.closed = False

        self._name = self.name
        if self.snapname:
            self._name += "@" + self.snapname

        if self._name not in ioctx.rbd_images:
            raise rbd.ImageNotFound(self.name)

    def size(self):
        assert not self.closed
        return self.ioctx.rbd_images[self.name]["size"]

    def resize(self, size):
        assert not self.closed
        self.ioctx.rbd_images[self.name]["size"] = size

    def lock_exclusive(self, cookie):
        assert not self.closed
        lock = self.ioctx.rbd_images[self.name]["lock"]
        if lock is None:
            self.ioctx.rbd_images[self.name]["lock"] = {
                "tag": None,
                "exclusive": True,
                "lockers": [("client.xyz", cookie, "127.0.0.1:9999")],
            }
            return
        else:
            assert lock["exclusive"]
            if not lock["lockers"][0][1] == cookie:
                raise rbd.ImageBusy(errno.EBUSY, "Image is busy")
            return
        raise RuntimeError("unsupported mock path")

    def lock_shared(self, cookie, tag):
        assert not self.closed
        lock = self.ioctx.rbd_images[self.name]["lock"]
        if lock is None:
            self.ioctx.rbd_images[self.name]["lock"] = {
                "tag": tag,
                "exclusive": False,
                "lockers": [("client.xyz", cookie, "127.0.0.1:9999")],
            }
            return
        else:
            if lock["exclusive"]:
                raise rbd.ImageBusy("already exclusively locked")
            if lock["tag"] != tag:
                raise rbd.ImageBusy("wrong tag")
            for l_client, l_cookie, l_addr in list(lock["lockers"]):
                if l_cookie != cookie:
                    lock["lockers"].append(
                        ("client.xyz", cookie, "127.0.0.1:9999")
                    )
                else:
                    raise rbd.ImageExists()
            # XXX we every only calls this from the same host so we never
            # actually get multiple lockers, just valid noops.
            return
        raise RuntimeError("unsupported mock path")

    def list_lockers(self):
        assert not self.closed
        if self.name not in self.ioctx.rbd_images:
            raise rbd.ImageNotFound(self.name)
        lock = self.ioctx.rbd_images[self.name]["lock"]
        return [] if lock is None else lock

    def list_snaps(self):
        assert not self.closed
        result = []
        for image, data in list(self.ioctx.rbd_images.items()):
            if image.startswith(self.name + "@"):
                snap = {
                    "id": data["snapid"],
                    "size": data["size"],
                    "name": image.split("@")[1],
                }
                result.append(snap)
        return result

    def create_snap(self, snapname):
        self.ioctx._rbd_create_snap(self.name, snapname)

    def remove_snap(self, snapname):
        self.ioctx._rbd_remove_snap(self.name, snapname)

    def unlock(self, cookie):
        assert not self.closed
        lock = self.ioctx.rbd_images[self.name]["lock"]
        if lock:
            assert lock["lockers"][0][1] == cookie
            self.ioctx.rbd_images[self.name]["lock"] = None

    def break_lock(self, client, cookie):
        assert not self.closed
        lock = self.ioctx.rbd_images[self.name]["lock"]
        if lock:
            assert lock["lockers"][0][0] == client
            assert lock["lockers"][0][1] == cookie
        self.ioctx.rbd_images[self.name]["lock"] = None

    def close(self):
        self.closed = True


@pytest.fixture
def ceph_mock(request, monkeypatch, tmpdir):
    is_live = request.node.get_closest_marker("live")
    if is_live is not None:
        # This is a live test. Do not mock things.
        return

    def image_map(self):
        if self.device:
            return
        self.device = str(tmpdir / self.fullname.replace("/", "-"))

        raw = self.device + ".raw"
        if not os.path.exists(raw):
            with open(raw, "wb") as f:
                f.seek(self.size)
                f.write(b"\0")
                f.close()
        os.symlink(raw, self.device)

        # create an implicit first partition as we can't really do the
        # partprobe dance.
        raw = self.part1dev + ".raw"
        with open(raw, "wb") as f:
            f.seek(self.size)
            f.write(b"\0")
            f.close()
        os.symlink(raw, self.part1dev)

    def image_unmap(self):
        if self.device is None:
            return
        os.unlink(self.device)
        os.unlink(self.part1dev)
        self.device = None

    monkeypatch.setattr(rados, "Rados", RadosMock)
    monkeypatch.setattr(rbd, "RBD", RBDMock)
    monkeypatch.setattr(rbd, "Image", ImageMock)
    monkeypatch.setattr(volume.Image, "map", image_map)
    monkeypatch.setattr(volume.Image, "unmap", image_unmap)


@pytest.fixture
def ceph_inst(ceph_mock):
    cfg = {
        "resource_group": "test",
        "rbd_pool": "rbd.hdd",
        "name": "simplevm",
        "disk": 10,
    }
    ceph = Ceph(cfg)
    ceph.CREATE_VM = "echo {name}"
    ceph.MKFS_XFS = "-q -f -K"
    ceph.__enter__()
    try:
        yield ceph
    finally:
        ceph.__exit__(None, None, None)



import contextlib
import copy
import datetime
import distutils.spawn
import fcntl
import glob
import json
import math
import os
import os.path as p
import shutil
import socket
import subprocess
import sys
import time
import typing
from codecs import decode
from multiprocessing.pool import ThreadPool

import colorama
import consulate
import consulate.models.agent
import pkg_resources
import requests
import yaml

from . import directory, util
from .exc import (
    ConfigChanged,
    InvalidCommand,
    VMConfigNotFound,
    VMStateInconsistent,
)
from .hazmat.ceph import Ceph
from .hazmat.cpuscan import scan_cpus
from .hazmat.qemu import Qemu, detect_current_machine_type
from .incoming import IncomingServer
from .outgoing import Outgoing
from .sysconfig import sysconfig
from .timeout import TimeOut
from .util import GiB, MiB, locate_live_service, log


def _handle_consul_event(event):
    handler = ConsulEventHandler()
    handler.handle(event)


OK, WARNING, CRITICAL, UNKNOWN = 0, 1, 2, 3

EXECUTABLE = sys.argv[0]  # make mockable


def unwrap_consul_armour(value: str) -> object:
    # See test_consul.py:prepare_consul_event.
    # The directory is providing a somewhat weird (likely historical)
    # structure with an ASCII armour. Due to Python only encoding/decoding
    # base64 on binary strings we need to do this little dance here.
    value = value.encode("ascii")
    value = decode(value, "base64")
    value = value.decode("ascii")
    return json.loads(value)


class ConsulEventHandler(object):
    """This is a wrapper that processes various consul events and multiplexes
    them into individual method handlers.

    """

    def handle(self, event):
        """Actual handling of a single Consul event in a
        separate process."""
        try:
            log.debug("handle-key", key=event["Key"])
            prefix = event["Key"].split("/")[0]
            if not event["Value"]:
                log.debug("ignore-key", key=event["Key"], reason="empty value")
                return
            getattr(self, prefix)(event)
        except BaseException as e:  # noqa
            # This must be a bare-except as it protects threads and the main
            # loop from dying. It could be that I'm wrong, but I'm leaving this
            # in for good measure.
            log.exception(
                "handle-key-failed", key=event.get("Key", None), exc_info=True
            )
        log.debug("finish-handle-key", key=event.get("Key", None))

    def node(self, event):
        config = unwrap_consul_armour(event["Value"])

        config["consul-generation"] = event["ModifyIndex"]
        vm = config["name"]
        if config["parameters"]["machine"] != "virtual":
            log.debug(
                "ignore-consul-event",
                machine=vm,
                reason="is a physical machine",
            )
            return
        log_ = log.bind(machine=vm)
        agent = Agent(vm, config)
        log_.info("processing-consul-event", consul_event="node")
        if agent.stage_new_config():
            cmd = [EXECUTABLE, "-D", "ensure", vm]
            log_.info("launch-ensure", cmd=cmd)
            s = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                close_fds=True,
            )
            log_.debug("launch-ensure", subprocess_pid=s.pid)
            stdout, stderr = s.communicate()
            exit_code = s.wait()
            log_.debug("launch-ensure", exit_code=exit_code)
            if exit_code:
                # Avoid logging things twice. However, if it failed then
                # we might be missing output that was generated before setting
                # up logging.
                log_.debug("launch-ensure", output=stdout)
        else:
            log_.info(
                "ignore-consul-event", machine=vm, reason="config is unchanged"
            )

    def snapshot(self, event):
        value = unwrap_consul_armour(event["Value"])
        vm = value["vm"]
        snapshot = value["snapshot"]
        log_ = log.bind(snapshot=snapshot, machine=vm)
        try:
            agent = Agent(vm)
        except RuntimeError:
            log_.warning("snapshot-ignore", reason="failed loading config")
            return
        with agent:
            log_.info("snapshot")
            try:
                agent.snapshot(snapshot, keep=0)
            except InvalidCommand:
                # The VM isn't in a state to make a snapshot. This is important
                # information for regular users but not for consul - ignoring
                # it is the right choice here. This will happen regularly
                # when the VM is running on a different host.
                pass


def running(expected=True):
    def wrap(f):
        def checked(self, *args, **kw):
            if self.qemu.is_running() != expected:
                self.log.error(
                    f.__name__,
                    expected="VM {}running".format("" if expected else "not "),
                )
                raise InvalidCommand(
                    "Invalid command: VM must {}be running to use `{}`.".format(
                        "" if expected else "not ", f.__name__
                    )
                )
            return f(self, *args, **kw)

        return checked

    return wrap


def locked(blocking=True):
    # This is thread-safe *AS LONG* as every thread uses a separate instance
    # of the agent. Using multiple file descriptors will guarantee that the
    # lock can only be held once even within a single process.
    #
    # However, we ensure locking status even for re-entrant / recursive usage.
    # For that we keep a counter how often we successfully acquired the lock
    # and then unlock when we're back to zero.
    def lock_decorator(f):
        def locked_func(self, *args, **kw):
            # New lockfile behaviour: lock with a global file that is really
            # only used for this purpose and is never replaced.
            self.log.debug("acquire-lock", target=self.lockfile)
            if not self._lockfile_fd:
                if not os.path.exists(self.lockfile):
                    open(self.lockfile, "a+").close()
                self._lockfile_fd = os.open(self.lockfile, os.O_RDONLY)
            mode = fcntl.LOCK_EX | (fcntl.LOCK_NB if not blocking else 0)
            try:
                fcntl.flock(self._lockfile_fd, mode)
            except IOError:
                # This happens in nonblocking mode and we just give up as
                # that's what's expected to speed up things.
                self.log.info(
                    "acquire-lock",
                    result="failed",
                    action="exit",
                    mode="nonblocking",
                )
                return os.EX_TEMPFAIL
            self.log.debug(
                "acquire-lock", target=self.lockfile, result="locked"
            )

            self._lock_count += 1
            self.log.debug("lock-status", count=self._lock_count)
            try:
                return f(self, *args, **kw)
            finally:
                self._lock_count -= 1
                self.log.debug("lock-status", count=self._lock_count)

                if self._lock_count == 0:
                    # New
                    try:
                        self.log.debug("release-lock", target=self.lockfile)
                        fcntl.flock(self._lockfile_fd, fcntl.LOCK_UN)
                        self.log.debug(
                            "release-lock",
                            target=self.lockfile,
                            result="unlocked",
                        )
                    except:  # noqa
                        self.log.debug(
                            "release-lock",
                            exc_info=True,
                            target=self.lockfile,
                            result="error",
                        )
                        pass

        return locked_func

    return lock_decorator


def swap_size(memory):
    """Returns the swap partition size in bytes."""
    swap_mib = max(1024, 32 * math.sqrt(memory))
    return int(swap_mib) * MiB


def tmp_size(disk):
    """Returns the tmp partition size in bytes."""
    tmp_gib = max(5, math.sqrt(disk))
    return int(tmp_gib) * GiB


class Agent(object):
    """The agent to control a single VM."""

    # Attributes on this class can be overriden (in a controlled fashion
    # from the sysconfig module. See this class' __init__. The defaults
    # are here to support testing.

    prefix = "/"
    this_host = ""
    migration_ctl_address = None
    accelerator = ""
    machine_type = "pc-i440fx"
    vhost = False
    ceph_id = "admin"
    timeout_graceful = 10
    vm_config_template_path = [
        "/etc/qemu/qemu.vm.cfg.in",
        pkg_resources.resource_filename(__name__, "qemu.vm.cfg.in"),
    ]
    consul_token = None
    consul_generation = -1

    # The binary generation is used to signal into a VM that the host
    # environment has changed in a way that requires a _cold_ reboot, thus
    # asking the VM to shut down cleanly so we can start it again. Change this
    # introducing a new Qemu machine type, or providing security fixes that
    # can't be applied by live migration.
    # The counter should be increased through the platform management as
    # this may change independently from fc.qemu releases.
    binary_generation = 0

    # For upgrade-purposes we're running an old and a new locking mechanism
    # in step-lock. We used to lock the configfile but we're using rename to
    # update it atomically. That's not compatible and will result in a consul
    # event replacing the file and then getting a lock on the new file while
    # there still is another process running. This will result in problems
    # accessing the QMP socket as that only accepts a single connection and
    # will then time out.
    _lock_count = 0
    _lockfile_fd = None
    _configfile_fd = None

    def __init__(self, name, enc=None):
        # Update configuration values from system or test config.
        self.log = log.bind(machine=name)

        self.__dict__.update(sysconfig.agent)

        self.name = name

        # The ENC data is the parsed configuration from the config file, or
        # from consul (or from wherever). This is intended to stay true to
        # the config file format (not adding computed values or manipulating
        # them) so that we can generate a new config file based on this data
        # through `stage_new_config()`.
        if enc is not None:
            self.enc = enc
        else:
            self.enc = self._load_enc()

        self.consul = consulate.Consul(token=self.consul_token)
        self.contexts = ()

    @property
    def configfile(self):
        return os.path.join(
            self.prefix, "etc/qemu/vm", "{}.cfg".format(self.name)
        )

    @property
    def configfile_staging(self):
        return os.path.join(
            self.prefix, "etc/qemu/vm", ".{}.cfg.staging".format(self.name)
        )

    @property
    def lockfile(self):
        return os.path.join(
            self.prefix, "run", "qemu.{}.lock".format(self.name)
        )

    def _load_enc(self):
        try:
            with open(self.configfile) as f:
                return yaml.safe_load(f)
        except IOError:
            if os.path.exists(self.configfile_staging):
                # The VM has been freshly created. Set up a boilerplate
                # configuration and let the actual config be established
                # through the 'ensure' command.
                return None
            else:
                self.log.error("unknown-vm")
                raise VMConfigNotFound(
                    "Could not load {}".format(self.configfile)
                )

    @classmethod
    def handle_consul_event(cls, input: typing.TextIO = sys.stdin):
        # Python in our environments defaults to UTF-8 and we generally
        # should be fine expecting a TextIO here as json load also expects
        # text input. It might be necessary at some point to explicitly
        # expect BinaryIO input and then enforce UTF-8 at this point.
        events = json.load(input)
        if not events:
            return
        log.info("start-consul-events", count=len(events))
        pool = ThreadPool(sysconfig.agent.get("consul_event_threads", 3))
        for event in sorted(events, key=lambda e: e.get("Key")):
            pool.apply_async(_handle_consul_event, (event,))
            time.sleep(0.05)
        pool.close()
        pool.join()
        log.info("finish-consul-events")

    @classmethod
    def _vm_agents_for_host(cls):
        for candidate in sorted(glob.glob(cls.prefix + "run/qemu.*.pid")):
            name = candidate.replace(cls.prefix + "run/qemu.", "")
            name = name.replace(".pid", "")
            try:
                agent = Agent(name)
                yield agent
            except Exception:
                log.exception("load-agent", machine=name, exc_info=True)

    @classmethod
    def report_supported_cpu_models(self):
        variations = []
        for variation in scan_cpus():
            log.info(
                "supported-cpu-model",
                architecture=variation.model.architecture,
                id=variation.cpu_arg,
                description=variation.model.description,
            )
            variations.append(variation.cpu_arg)

        d = directory.connect()
        d.report_supported_cpu_models(variations)

    @classmethod
    def ls(cls):
        """List all VMs that this host knows about.

        This means specifically that we have status (PID) files for the
        VM in the expected places.

        Gives a quick status for each VM.
        """
        for vm in cls._vm_agents_for_host():
            with vm:
                running = vm.qemu.process_exists()

                if running:
                    vm_mem = vm.qemu.proc().memory_full_info()

                    expected_size = (
                        vm.cfg["memory"] * 1024 * 1024
                        + vm.qemu.vm_expected_overhead * 1024 * 1024
                    )

                    log.info(
                        "online",
                        machine=vm.name,
                        cores=vm.cfg["cores"],
                        memory_booked="{:,.0f}".format(vm.cfg["memory"]),
                        memory_pss="{:,.0f}".format(vm_mem.pss / MiB),
                        memory_swap="{:,.0f}".format(vm_mem.swap / MiB),
                    )
                else:
                    log.info("offline", machine=vm.name)

    @classmethod
    def maintenance_enter(cls):
        """Prepare the host for maintenance mode.

        process exit codes signal success or (temporary) failure
        """
        d = directory.connect()
        host = socket.gethostname()
        for attempt in range(3):
            log.info("request-evacuation")
            evacuated = d.evacuate_vms(host)
            if not evacuated:
                # need to call evacuate_vms again to arrive at the empty set
                break
            log.info("evacuation-started", vms=evacuated)
            time.sleep(5)

        log.info("evacuation-pending")
        # Trigger a gratuitous event handling cycle to help speed up the
        # migration.
        subprocess.call(["systemctl", "reload", "consul"])

        # Monitor whether there are still VMs running.
        timeout = TimeOut(300, interval=3)
        while timeout.tick():
            p = subprocess.Popen(
                ["pgrep", "-f", "qemu-system-x86_64"], stdout=subprocess.PIPE
            )
            p.wait()
            num_procs = len(p.stdout.read().splitlines())
            log.info(
                "evacuation-running",
                vms=num_procs,
                timeout_remaining=timeout.remaining,
            )
            if num_procs == 0:
                # We made it: no VMs remaining, so we can proceed with the
                # maintenance.
                log.info("evacuation-success")
                sys.exit(0)
            time.sleep(10)

        log.info("evacuation-timeout", action="postpone maintenance")
        sys.exit(75)

    @classmethod
    def shutdown_all(cls):
        """Shut down all VMs cleanly.

        Runs the shutdowns in parallel to speed up host reboots.
        """
        vms = []

        for vm in cls._vm_agents_for_host():
            with vm:
                running = vm.qemu.process_exists()
                if running:
                    vms.append(vm)

        log.info("shutdown-all", count=len(vms))

        if not vms:
            return

        if sys.stdout.isatty():
            print(
                colorama.Fore.RED,
                "The following VMs will be shut down: ",
                colorama.Style.RESET_ALL,
            )
            print("\t" + ",".join(vm.name for vm in vms))
            print()
            while True:
                choice = input(
                    colorama.Style.BRIGHT
                    + colorama.Fore.RED
                    + "[DANGER] Are you really sure to shut down all {} VMs? (yes/no)\n".format(
                        len(vms)
                    )
                    + colorama.Style.RESET_ALL
                )
                if choice == "yes":
                    break
                elif choice == "no":
                    return
                else:
                    print("Please type 'yes' or 'no'.")

        def stop_vm(vm):
            # Isolate the stop call into separate fc-qemu
            # processes to ensure reliability.
            log.info("shutdown", vm=vm.name)
            p = subprocess.Popen([EXECUTABLE, "stop", vm.name])
            p.wait()

        pool = ThreadPool(5)
        for vm in vms:
            pool.apply_async(stop_vm, (vm,))
            time.sleep(0.5)
        pool.close()
        pool.join()
        log.info("shutdown-all", result="finished")

    @classmethod
    def check(cls):
        """Perform a health check of this host from a Qemu perspective.

        Checks:

        * no VM exceeds their PSS by guest memory + 2x expected overhead
        * VMs have very little swap (<1 GiB)
        * the total of the VMs PSS does not exceed
          total of guest + expected overhead

        Sets exit code according to the Nagios specification.

        """
        vms = list(cls._vm_agents_for_host())

        large_overhead_vms = []
        swapping_vms = []
        total_guest_and_overhead = 0
        expected_guest_and_overhead = 0

        # individual VMs ok?
        for vm in vms:
            with vm:
                try:
                    vm_mem = vm.qemu.proc().memory_full_info()
                except Exception:
                    # It's likely that the process went away while we analyzed
                    # it. Ignore.
                    continue
                if vm_mem.swap > 1 * GiB:
                    swapping_vms.append(vm)
                expected_size = (
                    vm.cfg["memory"] * MiB
                    + 2 * vm.qemu.vm_expected_overhead * MiB
                )
                expected_guest_and_overhead += (
                    vm.cfg["memory"] * MiB + vm.qemu.vm_expected_overhead * MiB
                )
                total_guest_and_overhead += vm_mem.pss
                if vm_mem.pss > expected_size:
                    large_overhead_vms.append(vm)

        output = []
        result = OK
        if large_overhead_vms:
            result = WARNING
            output.append(
                "VMs with large overhead: "
                + ",".join(x.name for x in large_overhead_vms)
            )
        if swapping_vms:
            result = WARNING
            output.append(
                "VMs swapping:" + ",".join(x.name for x in swapping_vms)
            )
        if total_guest_and_overhead > expected_guest_and_overhead:
            result = CRITICAL
            output.append("High total overhead")

        if result is OK:
            output.insert(0, "OK")
        elif result is WARNING:
            output.insert(0, "WARNING")
        elif result is CRITICAL:
            output.insert(0, "CRITICAL")
        else:
            output.insert(0, "UNKNOWN")

        output.insert(1, "{} VMs".format(len(vms)))
        output.insert(
            2, "{:,.0f} MiB used".format(total_guest_and_overhead / MiB)
        )
        output.insert(
            3, "{:,.0f} MiB expected".format(expected_guest_and_overhead / MiB)
        )

        print(" - ".join(output))

        return result

    def stage_new_config(self):
        """Save the current config on the agent into a staging config file.

        This method is safe to call outside of the agent's context manager.

        """
        # Ensure the config directory exists
        if not p.isdir(p.dirname(self.configfile)):
            os.makedirs(p.dirname(self.configfile))

        # Ensure the staging file exists.
        open(self.configfile_staging, "a+").close()
        staging_lock = os.open(self.configfile_staging, os.O_RDONLY)
        fcntl.flock(staging_lock, fcntl.LOCK_EX)
        self.log.debug(
            "acquire-staging-lock",
            target=self.configfile_staging,
            result="locked",
        )
        try:
            update_staging_config = False
            # Verify generation of config to protect against lost updates.
            with open(self.configfile_staging, "r") as current_staging:
                try:
                    current_staging_config = yaml.safe_load(current_staging)
                    current_staging_generation = current_staging_config[
                        "consul-generation"
                    ]
                except Exception:
                    self.log.debug("inconsistent-staging-config")
                    # Inconsistent staging configs should be updated
                    update_staging_config = True
                else:
                    # Newer staging configs should be updated
                    update_staging_config = (
                        self.enc["consul-generation"]
                        > current_staging_generation
                    )

            if update_staging_config:
                self.log.debug("save-staging-config")
                # Update the file in place to avoid lock breaking.
                with open(self.configfile_staging, "w") as new_staging:
                    yaml.safe_dump(self.enc, new_staging)
                os.chmod(self.configfile_staging, 0o644)

            # Do we need to activate this config?
            activate_staging_config = False
            try:
                with open(self.configfile, "r") as current_active:
                    current_active_config = yaml.safe_load(current_active)
                current_active_generation = current_active_config[
                    "consul-generation"
                ]
            except Exception:
                self.log.debug("inconsistent-active-config")
                # Inconsistent staging configs should be updated
                activate_staging_config = True
            else:
                # Newer staging configs should be updated
                activate_staging_config = (
                    self.enc["consul-generation"] > current_active_generation
                )
            return activate_staging_config
        finally:
            fcntl.flock(staging_lock, fcntl.LOCK_UN)
            os.close(staging_lock)
            self.log.debug(
                "release-staging-lock",
                target=self.configfile_staging,
                result="released",
            )

    @locked()
    def activate_new_config(self):
        """Activate the current staged config (if any).

        After calling this method, the agent's context manager must be
        re-entered to activate the changed configuration.

        """
        if not os.path.exists(self.configfile_staging):
            self.log.debug("check-staging-config", result="none")
            return False

        staging_lock = os.open(self.configfile_staging, os.O_RDONLY)
        fcntl.flock(staging_lock, fcntl.LOCK_EX)
        self.log.debug(
            "acquire-staging-lock",
            target=self.configfile_staging,
            result="locked",
        )
        try:
            # Verify generation of config to protect against lost updates.
            with open(self.configfile_staging, "r") as current_staging:
                try:
                    current_staging_config = yaml.safe_load(current_staging)
                    staging_generation = current_staging_config[
                        "consul-generation"
                    ]
                except Exception:
                    self.log.debug(
                        "update-check", result="inconsistent", action="purge"
                    )
                    os.unlink(self.configfile_staging)
                    return False
                else:
                    if staging_generation <= self.consul_generation:
                        # Stop right here, do not write a new config if the
                        # existing one is newer (or as new) already.
                        self.log.debug(
                            "update-check",
                            result="stale-update",
                            action="ignore",
                            update=staging_generation,
                            current=self.consul_generation,
                        )
                        # The old staging file needs to stay around so that
                        # the consul writer knows whether to launch an ensure
                        # agent or not.
                        return False
                self.log.debug(
                    "update-check",
                    result="update-available",
                    action="update",
                    update=staging_generation,
                    current=self.consul_generation,
                )

            # The config seems consistent and newer, lets update.
            # We can replace the config file because that one is protected
            # by the global VM lock.
            shutil.copy2(self.configfile_staging, self.configfile)
            self.enc = self._load_enc()
            return True
        finally:
            fcntl.flock(staging_lock, fcntl.LOCK_UN)
            os.close(staging_lock)
            self.log.debug(
                "release-staging-lock",
                target=self.configfile_staging,
                result="released",
            )

    def has_new_config(self):
        if not os.path.exists(self.configfile_staging):
            self.log.debug("check-staging-config", result="none")
            return False

        staging_lock = os.open(self.configfile_staging, os.O_RDONLY)
        fcntl.flock(staging_lock, fcntl.LOCK_EX)
        self.log.debug(
            "acquire-staging-lock",
            target=self.configfile_staging,
            result="locked",
        )
        try:
            # Verify generation of config to protect against lost updates.
            with open(self.configfile_staging, "r") as current_staging:
                try:
                    current_staging_config = yaml.safe_load(current_staging)
                    staging_generation = current_staging_config[
                        "consul-generation"
                    ]
                except Exception:
                    self.log.debug(
                        "update-check", result="inconsistent", action="purge"
                    )
                    os.unlink(self.configfile_staging)
                    return False
                else:
                    if staging_generation <= self.consul_generation:
                        # Stop right here, do not write a new config if the
                        # existing one is newer (or as new) already.
                        self.log.debug(
                            "update-check",
                            result="stale-update",
                            action="ignore",
                            update=staging_generation,
                            current=self.consul_generation,
                        )
                        # The old staging file needs to stay around so that
                        # the consul writer knows whether to launch an ensure
                        # agent or not.
                        return False
                self.log.debug(
                    "update-check",
                    result="update-available",
                    action="update",
                    update=staging_generation,
                    current=self.consul_generation,
                )
            return True
        finally:
            fcntl.flock(staging_lock, fcntl.LOCK_UN)
            os.close(staging_lock)
            self.log.debug(
                "release-staging-lock",
                target=self.configfile_staging,
                result="released",
            )

    def _update_from_enc(self):
        # This copy means we can't manipulate `self.cfg` to update ENC data,
        # which is OK. We did this at some point and we're adding computed
        # data to the `cfg` structure, that we do not want to accidentally
        # reflect back into the config file.
        self.cfg = copy.copy(self.enc["parameters"])
        self.cfg["name"] = self.enc["name"]
        self.cfg["swap_size"] = swap_size(self.cfg["memory"])
        self.cfg["tmp_size"] = tmp_size(self.cfg["disk"])
        self.cfg["ceph_id"] = self.ceph_id
        self.cfg["cpu_model"] = self.cfg.get("cpu_model", "qemu64")
        self.consul_generation = self.enc["consul-generation"]
        self.qemu = Qemu(self.cfg)
        self.ceph = Ceph(self.cfg)
        self.contexts = [self.qemu, self.ceph]
        for attr in ["migration_ctl_address"]:
            setattr(self, attr, getattr(self, attr).format(**self.cfg))
        for cand in self.vm_config_template_path:
            if p.exists(cand):
                self.vm_config_template = cand
                break

    def __enter__(self):
        # Allow updating our config by exiting/entering after setting new ENC
        # data.
        if self.enc is None:
            return
        self._update_from_enc()
        for c in self.contexts:
            c.__enter__()

    def __exit__(self, exc_value, exc_type, exc_tb):
        for c in self.contexts:
            try:
                c.__exit__(exc_value, exc_type, exc_tb)
            except Exception:
                self.log.exception("leave-subsystems", exc_info=True)

    def ensure(self):
        # Run the config activation code at least once, but then only if there
        # are still config changes around. We do this because multiple updates
        # may be piling up and we want to catch up as quickly as possible but
        # we only stage the updates and don't want to spawn an agent for each
        # update of each VM. So if an agent is already running it checks again
        # after it did something and then keeps going. OTOH agents that are
        # spawned after staging a new config will give up immediately when the
        # notice another agent running.

        # The __exit__/__enter__ dance is to simplify handling this special
        # case where we have to handle this method differently as it keeps
        # entering/exiting while cycling through the config changes. This keeps
        # the API and unit tests a bit cleaner even though we enter/exit one
        # time too many.
        self.__exit__(None, None, None)
        try:
            first = True
            while self.has_new_config() or first:
                self.log.info(
                    "running-ensure", generation=self.consul_generation
                )
                first = False
                try:
                    locking_code = self.ensure_()
                    if locking_code == os.EX_TEMPFAIL:
                        # We didn't get the lock, so someone else is
                        # already around who will pick up the (additional)
                        # changed config later.
                        break
                except ConfigChanged:
                    # Well then. Let's try this again.
                    continue
        finally:
            self.__enter__()
        self.log.debug("changes-settled")

    # This needs to be non-blocking at least to allow triggering an `fc-qemu
    # ensure` after a VM has spontaneously exited in the `run_supervised` code
    # but maybe also due to other situations where we want to avoid
    # deadlocks.
    @locked(blocking=False)
    def ensure_(self):
        self.activate_new_config()
        with self:
            # Host assignment is a bit tricky: we decided to not interpret an
            # *empty* cfg['kvm_host'] as "should not be running here" for the
            # sake of not accidentally causing downtime.
            try:
                if not self.cfg["online"]:
                    # Wanted offline.
                    self.ensure_offline()

                elif self.cfg["kvm_host"]:
                    if self.cfg["kvm_host"] != self.this_host:
                        # Wanted online, but explicitly on a different host.
                        self.ensure_online_remote()
                    else:
                        # Wanted online, and it's OK if it's running here, but
                        # I'll only start it if it really is wanted here.
                        self.ensure_online_local()

                self.raise_if_inconsistent()
            except VMStateInconsistent:
                # Last-resort seat-belt to verify that we ended up in a
                # consistent state. Inconsistent states result in the VM being
                # forcefully terminated.
                self.log.error(
                    "inconsistent-state", action="destroy", exc_info=True
                )
                self.qemu.destroy()

    def ensure_offline(self):
        if self.qemu.is_running():
            self.log.info(
                "ensure-state", wanted="offline", found="online", action="stop"
            )
            self.stop()
        else:
            self.log.info(
                "ensure-state", wanted="offline", found="offline", action="none"
            )

    def ensure_online_remote(self):
        if not self.qemu.is_running():
            # This cleans up potential left-overs from a previous migration.
            self.ceph.stop()
            self.consul_deregister()
            self.cleanup()
            return

        migration_errors = self.outmigrate()
        if migration_errors:
            # Stop here, do not clean up. This is fine as we may not have
            # made contact with our destination host. We keep trying later.
            return

        # Ensure ongoing maintenance for a VM that was potentially outmigrated.
        # But where the outmigration may have left bits and pieces.
        self.ceph.stop()
        self.consul_deregister()
        self.cleanup()

    def ensure_online_local(self):
        # Ensure state
        agent_likely_ready = True
        if not self.qemu.is_running():
            current_host = self._requires_inmigrate_from()
            if current_host:
                self.log.info(
                    "ensure-state",
                    wanted="online",
                    found="offline",
                    action="inmigrate",
                    remote=current_host,
                )
                exitcode = self.inmigrate()
                if exitcode:
                    # This is suboptimal: I hate error returns,
                    # but the main method is also a command. If we did
                    # not succeed in migrating, then I also don't want the
                    # consul registration to happen.
                    return
            else:
                self.log.info(
                    "ensure-state",
                    wanted="online",
                    found="offline",
                    action="start",
                )
                self.start()
                agent_likely_ready = False
        else:
            self.log.info(
                "ensure-state", wanted="online", found="online", action=""
            )

        # Perform ongoing adjustments of the operational parameters of the
        # running VM.
        self.consul_register()
        self.ensure_online_disk_size()
        self.ensure_online_disk_throttle()
        self.ensure_watchdog()
        # Be aggressive/opportunistic about re-acquiring locks in case
        # they were taken away.
        self.ceph.lock()
        if agent_likely_ready:
            # This requires guest agent interaction and we should only
            # perform this when we haven't recently booted the machine to
            # reduce the time we're unnecessarily waiting for timeouts.
            self.ensure_thawed()
            self.mark_qemu_binary_generation()

    def cleanup(self):
        """Removes various run and tmp files."""
        self.qemu.clean_run_files()
        for tmp in glob.glob(self.configfile + "?*"):
            os.unlink(tmp)

    def ensure_thawed(self):
        self.log.info("ensure-thawed", volume="root")
        try:
            self.qemu.thaw()
        except Exception as e:
            self.log.error("ensure-thawed-failed", reason=str(e))

    def mark_qemu_binary_generation(self):
        self.log.info(
            "mark-qemu-binary-generation", generation=self.binary_generation
        )
        try:
            self.qemu.write_file(
                "/run/qemu-binary-generation-current",
                (str(self.binary_generation) + "\n").encode("ascii"),
            )
        except Exception as e:
            self.log.error("mark-qemu-binary-generation", reason=str(e))

    def ensure_online_disk_size(self):
        """Trigger block resize action for the root disk."""
        target_size = self.cfg["disk"] * (1024**3)
        if self.ceph.root.size >= target_size:
            self.log.info(
                "check-disk-size",
                wanted=target_size,
                found=self.ceph.root.size,
                action="none",
            )
            return
        self.log.info(
            "check-disk-size",
            wanted=target_size,
            found=self.ceph.root.size,
            action="resize",
        )
        self.qemu.resize_root(target_size)

    def ensure_online_disk_throttle(self):
        """Ensure throttling settings."""
        target = self.cfg.get(
            "iops", self.qemu.throttle_by_pool.get(self.cfg["rbd_pool"], 250)
        )
        devices = self.qemu.block_info()
        for device in list(devices.values()):
            current = device["inserted"]["iops"]
            if current != target:
                self.log.info(
                    "ensure-throttle",
                    device=device["device"],
                    target_iops=target,
                    current_iops=current,
                    action="throttle",
                )
                self.qemu.block_io_throttle(device["device"], target)
            else:
                self.log.info(
                    "ensure-throttle",
                    device=device["device"],
                    target_iops=target,
                    current_iops=current,
                    action="none",
                )

    def ensure_watchdog(self, action="none"):
        """Ensure watchdog settings."""
        self.log.info("ensure-watchdog", action=action)
        self.qemu.watchdog_action(action)

    @property
    def svc_name(self):
        """Consul service name."""
        return "qemu-{}".format(self.name)

    def consul_register(self):
        """Register running VM with Consul."""
        self.log.debug("consul-register")
        self.consul.agent.service.register(
            self.svc_name,
            address=self.this_host,
            check=consulate.models.agent.Check(
                name="qemu-process",
                args=[
                    "/bin/sh",
                    "-c",
                    "test -e /proc/$(< /run/qemu.{}.pid )/mem || exit 2".format(
                        self.name
                    ),
                ],
                interval="5s",
            ),
        )

    def consul_deregister(self):
        """De-register non-running VM with Consul."""
        try:
            if self.svc_name not in self.consul.agent.services():
                return
            self.log.info("consul-deregister")
            self.consul.agent.service.deregister("qemu-{}".format(self.name))
        except requests.exceptions.ConnectionError:
            pass
        except Exception:
            self.log.exception("consul-deregister-failed", exc_info=True)

    @locked()
    @running(False)
    def start(self):
        self.generate_config()
        self.ceph.start(self.enc, self.binary_generation)
        self.qemu.start()
        self.consul_register()
        self.ensure_online_disk_throttle()
        self.ensure_watchdog()
        # We exit here without releasing the ceph lock in error cases
        # because the start may have failed because of an already running
        # process. Removing the lock in that case is dangerous. OTOH leaving
        # the lock is never dangerous. Lets go with that.

    @contextlib.contextmanager
    def frozen_vm(self):
        """Ensure a VM has a non-changing filesystem, that is clean enough
        to be mounted with a straight-forward journal replay.

        This is implemented using the 'fs-freeze' API through the Qemu guest
        agent.

        As we can't indicate the cleanliness of a non-running
        (potentially crashed) VM we do only signal a successful freeze iff
        the freeze happened and succeeded.

        """
        frozen = False
        try:
            if self.qemu.is_running():
                self.log.info("freeze", volume="root")
                try:
                    self.qemu.freeze()
                    frozen = True
                except Exception as e:
                    self.log.error(
                        "freeze-failed",
                        reason=str(e),
                        action="continue",
                        machine=self.name,
                    )
            yield frozen
        finally:
            if self.qemu.is_running():
                self.ensure_thawed()

    @locked()
    @running(True)
    def snapshot(self, snapshot, keep=0):
        """Guarantees a _consistent_ snapshot to be created.

        If we can't properly freeze the VM then whoever needs a (consistent)
        snapshot needs to figure out whether to go forward with an
        inconsistent snapshot.

        """
        if keep:
            until = util.today() + datetime.timedelta(days=keep)
            snapshot = snapshot + "-keep-until-" + until.strftime("%Y%m%d")
        if snapshot in [x.snapname for x in self.ceph.root.snapshots]:
            self.log.info("snapshot-exists", snapshot=snapshot)
            return
        self.log.info("snapshot-create", name=snapshot)
        with self.frozen_vm() as frozen:
            if frozen:
                self.ceph.root.snapshots.create(snapshot)
            else:
                self.log.error("snapshot-ignore", reason="not frozen")
                raise RuntimeError("VM not frozen, not making snapshot.")

    @locked()
    def status(self):
        """Determine status of the VM."""
        if self.qemu.is_running():
            status = 0
            self.log.info("vm-status", result="online")
            for device in list(self.qemu.block_info().values()):
                self.log.info(
                    "disk-throttle",
                    device=device["device"],
                    iops=device["inserted"]["iops"],
                )
        else:
            status = 1
            self.log.info("vm-status", result="offline")
        for volume in self.ceph.volumes:
            locker = volume.lock_status()
            self.log.info("rbd-status", volume=volume.fullname, locker=locker)
        consul = locate_live_service(self.consul, "qemu-" + self.name)
        if consul:
            self.log.info(
                "consul", service=consul["Service"], address=consul["Address"]
            )
        else:
            self.log.info("consul", service="<not registered>")
        return status

    def telnet(self):
        """Open telnet connection to the VM monitor."""
        self.log.info("connect-via-telnet")
        telnet = distutils.spawn.find_executable("telnet")
        os.execv(telnet, ("telnet", "localhost", str(self.qemu.monitor_port)))

    @locked()
    @running(True)
    def stop(self):
        timeout = TimeOut(self.timeout_graceful, interval=3)
        self.log.info("graceful-shutdown")
        try:
            self.qemu.graceful_shutdown()
        except (socket.error, RuntimeError):
            pass
        while timeout.tick():
            self.log.debug("checking-offline", remaining=timeout.remaining)
            if not self.qemu.is_running():
                self.log.info("vm-offline")
                self.ceph.stop()
                self.consul_deregister()
                self.cleanup()
                self.log.info("graceful-shutdown-completed")
                break
        else:
            self.log.warn("graceful-shutdown-failed", reason="timeout")
            self.kill()

    @locked()
    def restart(self):
        self.log.info("restart-vm")
        self.stop()
        self.start()

    @locked()
    @running(True)
    def kill(self):
        self.log.info("kill-vm")
        timeout = TimeOut(15, interval=1, raise_on_timeout=True)
        self.qemu.destroy()
        while timeout.tick():
            if not self.qemu.is_running():
                self.log.info("killed-vm")
                self.ceph.stop()
                self.consul_deregister()
                self.cleanup()
                break
        else:
            self.log.warning("kill-vm-failed", note="Check lock consistency.")

    def _requires_inmigrate_from(self):
        """Check whether an inmigration makes sense.

        This makes sense if the VM isn't running locally and (Consul knows
        about a running VM or the VM is locked remotely).

        Returns the name of the remote host the VM should be migrated from
        or None if inmigration doesn't make sense.

        """
        existing = locate_live_service(self.consul, "qemu-" + self.name)

        if existing and existing["Address"] != self.this_host:
            # Consul knows about a running VM. Lets try a migration.
            return existing["Address"]

        if self.ceph.is_unlocked():
            # Consul doesn't know about a running VM and no volume is locked.
            # It doesn't make sense to live migrate this VM.
            return None

        if self.ceph.locked_by_me():
            # Consul doesn't know about a running VM and the volume is
            # locked by me, so it doesn't make sense to live migrate the VM.
            return None

        # The VM seems to be locked somewhere else, try to migrate it from
        # there.
        return self.ceph.locked_by()

    @locked()
    @running(False)
    def inmigrate(self):
        self.log.info("inmigrate")
        server = IncomingServer(self)
        exitcode = server.run()
        if not exitcode:
            self.consul_register()
        self.log.info("inmigrate-finished", exitcode=exitcode)
        return exitcode

    @locked()
    @running(True)
    def outmigrate(self):
        self.log.info("outmigrate")
        # re-register in case services got lost during Consul restart
        self.consul_register()
        client = Outgoing(self)
        exitcode = client()
        if not exitcode:
            self.consul_deregister()
        self.log.info("outmigrate-finished", exitcode=exitcode)
        return exitcode

    @locked()
    def lock(self):
        self.log.info("assume-all-locks")
        for vol in self.ceph.volumes:
            vol.lock()

    @locked()
    @running(False)
    def unlock(self):
        self.log.info("release-all-locks")
        self.ceph.stop()

    @running(False)
    def force_unlock(self):
        self.log.info("break-all-locks")
        self.ceph.force_unlock()

    # Helper methods
    def raise_if_inconsistent(self):
        """Raise an VMStateInconsistent error if the VM state is not
        consistent.

        Either all of the following or none of the following must be true:

        - The process belonging to the PID we know exists.
        - The QMP monitor reliably tells us that the Qemu status of the
          VM/CPUs is "running"
        - We have locked the volumes in Ceph.

        If they are not the same, then the state is considered inconsistent
        and this method will return False.

        """
        state = VMStateInconsistent()
        state.qemu = self.qemu.is_running()
        state.proc = bool(self.qemu.proc())
        state.ceph_lock = self.ceph.locked_by_me()
        self.log.debug(
            "check-state-consistency",
            is_consistent=state.is_consistent(),
            qemu=state.qemu,
            proc=state.proc,
            ceph_lock=state.ceph_lock,
        )
        if not state.is_consistent():
            raise state

    # CAREFUL: changing anything in this config files will require maintenance
    # w/ reboot of all VMs in the infrastructure. Need to increase the
    # binary generation for that, though.
    def generate_config(self):
        """Generate a new Qemu config (and options) for a freshly
        starting VM.

        The configs are intended to be non-host-specific.

        This two-step behaviour is needed to support migrating VMs
        and keeping arguments consistent while allowing to localize arguments
        to the host running or receiving the VM.

        """
        self.log.debug("generate-config")
        self.qemu.args = [
            "-nodefaults",
            "-only-migratable",
            "-cpu {cpu_model},enforce",
            # Watch out: kvm.name is used for sanity checking critical actions.
            "-name {name},process=kvm.{name}",
            "-chroot {{chroot}}",
            "-runas nobody",
            "-serial file:/var/log/vm/{name}.log",
            "-display vnc={{vnc}}",
            "-pidfile {{pidfile}}",
            "-vga std",
            # We use this '-m' flag to find what a running VM is actually
            # using at the moment. If this flag is changed then that code must
            # be adapted as well. This is used in incoming.py and qemu.py.
            "-m {memory}",
            "-readconfig {{configfile}}",
        ]
        self.qemu.args = [a.format(**self.cfg) for a in self.qemu.args]

        vhost = '  vhost = "on"' if self.vhost else ""

        netconfig = []
        for net, net_config in sorted(self.cfg["interfaces"].items()):
            ifname = "t{}{}".format(net, self.cfg["id"])
            netconfig.append(
                """
[device]
  driver = "virtio-net-pci"
  netdev = "{ifname}"
  mac = "{mac}"

[netdev "{ifname}"]
  type = "tap"
  ifname = "{ifname}"
  script = "/etc/kvm/kvm-ifup"
  downscript = "/etc/kvm/kvm-ifdown"
{vhost}
""".format(
                    ifname=ifname, mac=net_config["mac"], vhost=vhost
                )
            )

        with open(self.vm_config_template) as f:
            tpl = f.read()
        accelerator = (
            '  accel = "{}"'.format(self.accelerator)
            if self.accelerator
            else ""
        )
        machine_type = detect_current_machine_type(self.machine_type)
        self.qemu.config = tpl.format(
            accelerator=accelerator,
            machine_type=machine_type,
            disk_cache_mode=self.qemu.disk_cache_mode,
            network="".join(netconfig),
            **self.cfg,
        )



import os
import shutil
import subprocess
import sys
import traceback

import pkg_resources
import pytest
import structlog

import fc.qemu.agent
import fc.qemu.hazmat.qemu

from .agent import Agent


def pytest_collectstart(collector):
    from fc.qemu.sysconfig import sysconfig

    sysconfig.load_system_config()


@pytest.fixture(autouse=True)
def synthetic_root(request, monkeypatch, tmpdir):
    is_live = request.node.get_closest_marker("live")
    if is_live is not None:
        # This is a live test. Do not mock things.
        return
    monkeypatch.setattr(fc.qemu.hazmat.qemu.Qemu, "prefix", str(tmpdir))
    os.makedirs(str(tmpdir / "run"))
    os.makedirs(str(tmpdir / "etc/qemu/vm"))
    monkeypatch.setattr(Agent, "prefix", str(tmpdir))
    monkeypatch.setattr(fc.qemu.agent, "EXECUTABLE", "true")


@pytest.fixture(scope="session")
def setup_structlog():
    from . import util

    # set to True to temporarily get detailed tracebacks
    log_exceptions = False

    def test_logger(logger, method_name, event):
        stack = event.pop("stack", None)
        exc = event.pop("exception", None)
        event_name = event.pop("event", "")
        event_prefix = os.path.basename(event_name) if event_name else " "
        result = []
        if "output_line" in event:
            result = fc.qemu.logging.prefix(
                event_prefix, event["output_line"].strip()
            )
        else:
            output = event.pop("output", None)

            result = []
            if event_name:
                result.append(event_name)
            for key in sorted(event):
                result.append("{}={}".format(key, str(event[key]).strip()))
            result = " ".join(result)

            if output:
                result += fc.qemu.logging.prefix(event_prefix, output)

        # Ensure we get something to read on stdout in case we have errors.
        print(result)
        if stack:
            print(stack)
        if exc:
            print(exc)

        # Allow tests to inspect only methods and events they are interested
        # in. This reduces noise in our test outputs and comparisons and
        # reduces fragility.
        show_methods = util.test_log_options["show_methods"]
        if show_methods and method_name not in show_methods:
            raise structlog.DropEvent

        show_events = util.test_log_options["show_events"]
        if show_events:
            for show in show_events:
                if show in event_name:
                    break
            else:
                raise structlog.DropEvent

        util.log_data.append(result)
        if log_exceptions:
            if stack:
                util.log_data.extend(stack.splitlines())
            if exc:
                util.log_data.extend(exc.splitlines())
        raise structlog.DropEvent

    structlog.configure(
        processors=(
            ([structlog.processors.format_exc_info] if log_exceptions else [])
            + [test_logger]
        )
    )


@pytest.fixture(autouse=True)
def reset_structlog(setup_structlog):
    from . import util

    util.log_data = []
    util.test_log_options = {"show_methods": [], "show_events": []}


def pytest_assertrepr_compare(op, left, right):
    if left.__class__.__name__ == "Ellipsis":
        return left.compare(right).diff
    elif right.__class__.__name__ == "Ellipsis":
        return right.compare(left).diff


@pytest.fixture
def clean_environment():
    def clean():
        subprocess.call("pkill -f qemu", shell=True)
        subprocess.call("rbd rm rbd.ssd/simplevm.swap", shell=True)
        subprocess.call("rbd snap purge rbd.ssd/simplevm.root", shell=True)
        subprocess.call("rbd rm rbd.ssd/simplevm.root", shell=True)
        subprocess.call("rbd rm rbd.ssd/simplevm.tmp", shell=True)

    clean()
    yield
    clean()


@pytest.fixture
def vm(clean_environment, monkeypatch, tmpdir):
    import fc.qemu.hazmat.qemu

    monkeypatch.setattr(fc.qemu.hazmat.qemu.Qemu, "guestagent_timeout", 0.1)
    fixtures = pkg_resources.resource_filename(__name__, "tests/fixtures")
    shutil.copy(fixtures + "/simplevm.yaml", "/etc/qemu/vm/simplevm.cfg")
    if os.path.exists("/etc/qemu/vm/.simplevm.cfg.staging"):
        os.unlink("/etc/qemu/vm/.simplevm.cfg.staging")
    vm = Agent("simplevm")
    vm.timeout_graceful = 1
    vm.__enter__()
    vm.qemu.qmp_timeout = 0.1
    vm.qemu.vm_expected_overhead = 128
    for snapshot in vm.ceph.root.snapshots:
        snapshot.remove()
    vm.qemu.destroy()
    vm.unlock()
    get_log()
    yield vm

    for snapshot in vm.ceph.root.snapshots:
        snapshot.remove()
    exc_info = sys.exc_info()
    vm.__exit__(*exc_info)
    if len(exc_info):
        print(traceback.print_tb(exc_info[2]))


def get_log():
    from fc.qemu import util

    result = "\n".join(util.log_data)
    util.log_data = []
    return result



import pytest

from fc.qemu.util import log

from ..timeout import TimeOut, TimeoutError


def test_timeout_raises():
    log_ = log.bind(machine="test")
    timeout = TimeOut(0.1, raise_on_timeout=True, log=log_)
    with pytest.raises(TimeoutError):
        while timeout.tick():
            pass


def test_timeout_stops():
    log_ = log.bind(machine="test")
    timeout = TimeOut(0.1, log=log_)
    while timeout.tick():
        pass



import datetime
import shlex
import subprocess
import sys

from fc.qemu.main import daemonize
from fc.qemu.util import FlushingStream, ensure_separate_cgroup


def run_supervised(cmd, name, logfile):
    log = FlushingStream(open(logfile, "a+"))
    daemonize(log)
    now = datetime.datetime.now().isoformat()
    log.write("{} - starting command {}\n".format(now, cmd))
    s = subprocess.Popen(
        shlex.split(cmd), close_fds=True, stdin=None, stdout=log, stderr=log
    )
    now = datetime.datetime.now().isoformat()
    log.write("{} - command has PID {}\n".format(now, s.pid))
    exit_code = s.wait()
    now = datetime.datetime.now().isoformat()
    log.write("{} - command exited with exit code {}\n".format(now, exit_code))

    # Restart immediately using ensure. This can happen if a VM powers down
    # with the intention to get started with new settings or if qemu crashes.
    # If the VM really should be shut down, then fc-qemu won't do anything.
    # This requires that `agent.ensure` is using a non-blocking lock to
    # avoid deadlocks.
    log.write("{} - ensuring VM state".format(now))
    s = subprocess.Popen(
        ["fc-qemu", "-v", "ensure", name],
        close_fds=True,
        stdin=None,
        stdout=log,
        stderr=log,
        encoding="ascii",
        errors="replace",
    )
    exit_code = s.wait()
    now = datetime.datetime.now().isoformat()
    log.write(
        "{} - ensure command exited with exit code {}\n".format(now, exit_code)
    )


def main():
    ensure_separate_cgroup()
    run_supervised(*sys.argv[1:])



import itertools
import os
import subprocess

from fc.qemu.timeout import TimeOut, TimeoutError
from fc.qemu.util import log

from .qemu import Qemu

FNULL = open(os.devnull, "w")


class Model(object):

    architecture = None
    identifier = None
    description = None

    def __init__(self, architecture, identifier, description):
        self.architecture = architecture
        self.identifier = identifier
        self.description = description


class Variation(object):

    model = None
    flags = ()

    def __init__(self, model, flags):
        self.model = model
        self.flags = tuple(sorted(set(flags)))

    @property
    def cpu_arg(self):
        return ",".join((self.model.identifier,) + self.flags)


class QemuHost(object):
    @classmethod
    def detect(self):
        for line in open("/proc/cpuinfo"):
            if not line.startswith("vendor_id"):
                continue
            _, vendor = line.split(":")
            vendor = vendor.strip()
            for host in [AMDHost, IntelHost]:
                if host.vendor == vendor:
                    return host()
            break
        else:
            raise RuntimeError("Could not determine CPU vendor.")


class AbstractHost(QemuHost):

    CPU_MODELS = ["qemu64-v1"]
    BUG_FLAGS = []


class AMDHost(QemuHost):

    vendor = "AuthenticAMD"

    CPU_MODELS = [
        "qemu64-v1",
        "EPYC-v1",
        "EPYC-v2",
        "EPYC-v3",
        "EPYC-Rome-v1",
    ]

    BUG_FLAGS = [
        "ibpb",
        "virt-ssbd",
        "amd-ssbd",
        "amd-no-ssb",
        "pdpe1gb",
    ]


class IntelHost(QemuHost):

    vendor = "GenuineIntel"

    CPU_MODELS = [
        "Broadwell-v1",
        "Broadwell-v2",
        "Broadwell-v3",
        "Broadwell-v4",
        "Cascadelake-Server-v1",
        "Cascadelake-Server-v2",
        "Haswell-v1",
        "Haswell-v2",
        "Haswell-v3",
        "Haswell-v4",
        "IvyBridge-v1",
        "IvyBridge-v2",
        "Nehalem-v1",
        "Nehalem-v2",
        "SandyBridge-v1",
        "SandyBridge-v2",
        "Skylake-Server-v1",
        "Skylake-Server-v2",
        "Westmere-v1",
        "Westmere-v2",
        "qemu64-v1",
    ]

    BUG_FLAGS = [
        "pcid",
        "spec-ctrl",
        "ssbd",
        "pdpe1gb",
    ]


def scan_cpus(host=None):
    if host is None:
        host = QemuHost.detect()

    models = []
    for identifier in host.CPU_MODELS:
        models.append(Model("x86", identifier, ""))

    # Determine combinations with additional desirable flags
    desirable_flags = host.BUG_FLAGS
    desirable_combinations = []
    for L in range(0, len(desirable_flags) + 1):
        desirable_combinations.extend(
            itertools.combinations(desirable_flags, L)
        )

    variations = []

    for model in models:
        for combination in desirable_combinations:
            variations.append(Variation(model, combination))

    valid_models = []

    for variation in variations:
        log.debug(
            "test-cpu",
            id=variation.cpu_arg,
            description=variation.model.description,
            architecture=variation.model.architecture,
        )
        task = subprocess.Popen(
            [
                Qemu.executable,
                "-cpu",
                variation.cpu_arg + ",enforce",
                "-machine",
                "pc,accel=kvm",
                "-enable-kvm",
                "-monitor",
                "stdio",
                "-display",
                "none",
                "-nodefaults",
            ],
            stdin=subprocess.PIPE,
            stdout=FNULL,
            stderr=FNULL,
            encoding="ascii",
            errors="replace",
        )
        task.communicate(input="quit\n")
        if not task.wait():
            valid_models.append(variation)

    return valid_models



"""Low-level interface to Qemu commands."""

import datetime
import fcntl
import glob
import os.path
import socket
import subprocess
import time
from codecs import encode

import psutil
import yaml

from ..exc import QemuNotRunning, VMStateInconsistent
from ..sysconfig import sysconfig
from ..timeout import TimeOut
from ..util import ControlledRuntimeException, log
from . import supervise
from .guestagent import ClientError, GuestAgent
from .qmp import QEMUMonitorProtocol as Qmp
from .qmp import QMPConnectError

# Freeze requests may take a _long_ _long_ time and the default
# timeout of 3 seconds will cause everything to explode when
# the guest takes too long. We've seen 16 seconds as a regular
# period in some busy and large machines. I'm being _very_
# generous using a 120s timeout here.
# This is a global variable so we can instrument it during testing.
FREEZE_TIMEOUT = 120


class InvalidMigrationStatus(Exception):
    pass


def detect_current_machine_type(
    prefix: str, encoding="ascii", errors="replace"
):
    """Given a machine type prefix, e.g. 'pc-i440fx-' return the newest
    current machine on the available Qemu system.

    Newest in this case means the first item in the list as given by Qemu.
    """
    result = subprocess.check_output(
        [Qemu.executable, "-machine", "help"], encoding=encoding, errors=errors
    )
    for line in result.splitlines():
        if line.startswith(prefix):
            return line.split()[0]
    raise KeyError("No machine type found for prefix `{}`".format(prefix))


def locked_global(f):
    LOCK = "/run/fc-qemu.lock"
    # This is thread-safe *AS LONG* as every thread uses a separate instance
    # of the agent. Using multiple file descriptors will guarantee that the
    # lock can only be held once even within a single process.
    def locked(self, *args, **kw):
        self.log.debug("acquire-global-lock", target=LOCK)
        if not self._global_lock_fd:
            if not os.path.exists(LOCK):
                open(LOCK, "a+").close()
            self._global_lock_fd = os.open(LOCK, os.O_RDONLY)
        self.log.debug("global-lock-acquire", target=LOCK, result="locked")

        fcntl.flock(self._global_lock_fd, fcntl.LOCK_EX)
        self._global_lock_count += 1
        self.log.debug(
            "global-lock-status", target=LOCK, count=self._global_lock_count
        )
        try:
            return f(self, *args, **kw)
        finally:
            self._global_lock_count -= 1
            self.log.debug(
                "global-lock-status", target=LOCK, count=self._global_lock_count
            )
            if self._global_lock_count == 0:
                self.log.debug("global-lock-release", target=LOCK)
                fcntl.flock(self._global_lock_fd, fcntl.LOCK_UN)
                self.log.debug("global-lock-release", result="unlocked")

    return locked


class Qemu(object):

    prefix = ""
    executable = "qemu-system-x86_64"

    # Attributes on this class can be overriden (in a controlled fashion
    # from the sysconfig module. See this class' __init__. The defaults
    # are here to support testing.

    cfg = None
    require_kvm = True
    migration_address = None
    max_downtime = 1.0  # seconds
    # 0.8 * 10 Gbit/s in bytes/s
    migration_bandwidth = int(0.8 * 10 * 10**9 / 8)

    guestagent_timeout = 3.0
    # QMP runs in the main thread and can block. Our original 15s timeout
    # is definitely too short. Many discussions mention that 5 minutes have
    # stabilized the situation even under adverse situations.
    qmp_timeout = 5 * 60
    thaw_retry_timeout = 2
    vm_max_total_memory = 0  # MiB: maximum amount of booked memory (-m)
    # on this host
    vm_expected_overhead = 0  # MiB: expected amount of PSS overhead per VM

    # The non-hosts-specific config configuration of this Qemu instance.
    args = ()
    config = ""

    # Host-specific qemu configuration
    chroot = "/srv/vm/{name}"
    vnc = "localhost:1"

    MONITOR_OFFSET = 20000

    pidfile = "/run/qemu.{name}.pid"
    configfile = "/run/qemu.{name}.cfg"
    argfile = "/run/qemu.{name}.args"
    qmp_socket = "/run/qemu.{name}.qmp.sock"

    _global_lock_fd = None
    _global_lock_count = 0

    migration_lockfile = "/run/qemu.migration.lock"
    _migration_lock_fd = None

    def __init__(self, vm_cfg):
        # Update configuration values from system or test config.
        self.__dict__.update(sysconfig.qemu)

        self.cfg = vm_cfg
        # expand template keywords in configuration variables
        for f in [
            "pidfile",
            "configfile",
            "argfile",
            "migration_address",
            "qmp_socket",
        ]:
            setattr(self, f, getattr(self, f).format(**vm_cfg))

        # prepare qemu-specific config settings
        self.qemu_cfg = self.cfg.get("qemu", {})

        # The default if nothing is set is to enable "writeback" for backwards
        # compatability: before introducing this option everything used
        # writeback.
        if self.qemu_cfg.get("write_back_cache", True):
            self.disk_cache_mode = "writeback"
        else:
            self.disk_cache_mode = "none"

        # We are running qemu with chroot which causes us to not be able to
        # resolve names. :( See #13837.
        a = self.migration_address.split(":")
        if a[0] == "tcp":
            a[1] = socket.gethostbyname(a[1])
        self.migration_address = ":".join(a)
        self.name = self.cfg["name"]
        self.monitor_port = self.cfg["id"] + self.MONITOR_OFFSET
        self.guestagent = GuestAgent(self.name, timeout=self.guestagent_timeout)

        self.log = log.bind(machine=self.name, subsystem="qemu")

        self.pidfile = self.prefix + self.pidfile
        self.configfile = self.prefix + self.configfile
        self.argfile = self.prefix + self.argfile
        self.qmp_socket = self.prefix + self.qmp_socket
        self.migration_lockfile = self.prefix + self.migration_lockfile
        self.chroot = self.prefix + self.chroot

    __qmp = None

    @property
    def qmp(self):
        if self.__qmp is None:
            qmp = Qmp(self.qmp_socket, self.log)
            qmp.settimeout(self.qmp_timeout)
            try:
                qmp.connect()
            except socket.error:
                # We do not log this as this does happen quite regularly and
                # is usually fine as the VM wasn't started (yet).
                pass
            else:
                self.__qmp = qmp
        return self.__qmp

    def __enter__(self):
        pass

    def __exit__(self, exc_value, exc_type, exc_tb):
        if self.__qmp:
            self.__qmp.close()
            self.__qmp = None

    def proc(self):
        """Qemu processes as psutil.Process object.

        Returns None if the PID file does not exist or the process is
        not running.
        """
        try:
            with open(self.pidfile) as p:
                # pid file may contain trailing lines with garbage
                for line in p:
                    proc = psutil.Process(int(line))
                    marker = "{name},process=kvm.{name}".format(name=self.name)
                    # Do not use proc.name() here - it's only 16 bytes ...
                    if marker not in proc.cmdline():
                        break
                    return proc
        except (IOError, OSError, ValueError, psutil.NoSuchProcess):
            pass

    def prepare_log(self):
        if not os.path.exists("/var/log/vm"):
            raise RuntimeError("Expected directory /var/log/vm to exist.")
        logfile = "/var/log/vm/{}.log".format(self.name)
        alternate = "/var/log/vm/{}-{}.log".format(
            self.name, datetime.datetime.now().isoformat()
        )
        if os.path.exists(logfile):
            os.rename(logfile, alternate)

    def _current_vms_booked_memory(self):
        """Determine the amount of booked memory (MiB) from the

        currently running VM processes.
        """
        total = 0
        for proc in psutil.process_iter():
            try:
                pinfo = proc.as_dict(attrs=["pid", "name", "cmdline"])
            except psutil.NoSuchProcess:
                continue

            if not pinfo["name"].startswith("kvm."):
                continue
            if not (
                pinfo["cmdline"] and pinfo["cmdline"][0] == "qemu-system-x86_64"
            ):
                continue
            try:
                m_flag = pinfo["cmdline"].index("-m")
                memory = int(pinfo["cmdline"][m_flag + 1])
            except (ValueError, KeyError):
                self.log.debug(
                    "unexpected-cmdline",
                    cmdline=format(" ".join(pinfo["cmdline"])),
                )
                raise ControlledRuntimeException(
                    "Can not determine used memory for {}".format(
                        " ".join(pinfo["cmdline"])
                    )
                )
            total += memory + self.vm_expected_overhead
        return total

    def _verify_memory(self):
        """Verify that we do not accidentally run more VMs than we can
        physically bear.

        This is a protection to avoid starting new VMs while some old VMs
        that the directory assumes have been migrated or stopped already
        are still running. This can cause severe performance penalties and may
        also kill VMs under some circumstances.

        Also, if VMs should exhibit extreme overhead, we protect against
        starting additional VMs even if our inventory says we should be
        able to run them.

        If no limit is configured then we start VMs based on actual
        availability only.

        """
        current_booked = self._current_vms_booked_memory()  # MiB
        required = self.cfg["memory"] + self.vm_expected_overhead  # MiB

        available_real = psutil.virtual_memory().available / (1024 * 1024)
        limit_booked = self.vm_max_total_memory
        available_bookable = limit_booked - current_booked

        if (limit_booked and available_bookable < required) or (
            available_real < required
        ):
            self.log.error(
                "insufficient-host-memory",
                bookable=available_bookable,
                available=available_real,
                required=required,
            )
            raise ControlledRuntimeException(
                "Insufficient bookable memory to start VM."
            )

        self.log.debug(
            "sufficient-host-memory",
            bookable=available_bookable,
            available_real=available_real,
            required=required,
        )

    # This lock protects checking the amount of available memory and actually
    # starting the VM. This ensure that no other process checks at the same
    # time and we end up using the free memory twice.
    @locked_global
    def _start(self, additional_args=()):
        if self.require_kvm and not os.path.exists("/dev/kvm"):
            self.log.error("missing-kvm-support")
            raise ControlledRuntimeException(
                "Refusing to start without /dev/kvm support."
            )

        self._verify_memory()

        self.prepare_config()
        self.prepare_log()
        try:
            args = list(self.local_args) + list(additional_args)
            # We do not daemonize any longer and even want this to happen
            # when migrating VMs.
            # We also force the internal stderr log, even though I haven't
            # seen this being useful, yet.
            args = [x for x in args if x != "-daemonize"]
            args = [x for x in args if not x.startswith("-D ")]
            args.append("-D /var/log/vm/{}.qemu.internal.log".format(self.name))

            cmd = "{} {}".format(self.executable, " ".join(args))
            self.log.info("start-qemu")
            self.log.debug(
                self.executable,
                local_args=self.local_args,
                additional_args=additional_args,
            )
            # We explicitly close all fds for the child to avoid inheriting fd
            # locks accidentally and indefinitely.
            qemu_log = "/var/log/vm/{}.supervisor.log".format(self.name)
            cmdline = ["supervised-qemu", cmd, self.name, qemu_log]
            self.log.debug("exec", cmd=" ".join(cmdline))
            p = subprocess.Popen(
                cmdline,
                close_fds=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                encoding="ascii",
                errors="replace",
            )
            stdout, stderr = p.communicate()
            # FIXME: prettier logging
            self.log.debug("supervised-qemu-stdout", output=stdout)
            self.log.debug("supervised-qemu-stderr", output=stderr)
            if p.returncode != 0:
                raise QemuNotRunning(p.returncode, stdout, stderr)
        except QemuNotRunning:
            # Did not start. Not running.
            self.log.exception("qemu-failed")
            raise

    def start(self):
        self._start()
        timeout = TimeOut(10, 0.25, raise_on_timeout=True)
        while timeout.tick():
            if self.is_running():
                return

    def freeze(self):
        with self.guestagent as guest:
            try:
                # This request may take a _long_ _long_ time and the default
                # timeout of 3 seconds will cause everything to explode when
                # the guest takes too long. We've seen 16 seconds as a regular
                # period in some busy and large machines. I'm being _very_
                # generous using a 120s timeout here.
                guest.cmd("guest-fsfreeze-freeze", timeout=FREEZE_TIMEOUT)
            except ClientError:
                self.log.debug("guest-fsfreeze-freeze-failed", exc_info=True)
            assert guest.cmd("guest-fsfreeze-status") == "frozen"

    def _thaw_via_guest_agent(self):
        with self.guestagent as guest:
            try:
                guest.cmd("guest-fsfreeze-thaw")
            except ClientError:
                self.log.debug("guest-fsfreeze-freeze-thaw", exc_info=True)
                raise
            result = guest.cmd("guest-fsfreeze-status")
            if result != "thawed":
                raise RuntimeError("Unexpected thaw result: {}".format(result))

    def thaw(self):
        try:
            self._thaw_via_guest_agent()
        except Exception:
            self.log.warning("guest-fsfreeze-thaw-failed", exc_info=True)
            raise

    def write_file(self, path, content: bytes):
        if not isinstance(content, bytes):
            raise TypeError("Expected bytes, got string.")
        with self.guestagent as guest:
            try:
                handle = guest.cmd("guest-file-open", path=path, mode="w")
                guest.cmd(
                    "guest-file-write",
                    handle=handle,
                    **{"buf-b64": encode(content, "base64")},
                )
                guest.cmd("guest-file-close", handle=handle)
            except ClientError:
                self.log.error("guest-write-file", exc_info=True)

    def inmigrate(self):
        self._start(["-incoming {}".format(self.migration_address)])
        time.sleep(1)
        status = self.qmp.command("query-status")
        assert not status["running"], status
        assert status["status"] == "inmigrate", status
        return self.migration_address

    def migrate(self, address):
        """Initiate actual (out-)migration"""
        self.log.debug("migrate")
        self.qmp.command(
            "migrate-set-capabilities",
            capabilities=[
                {"capability": "xbzrle", "state": False},
                {"capability": "auto-converge", "state": True},
            ],
        )
        self.qmp.command(
            "migrate-set-parameters",
            **{
                "compress-level": 0,
                "downtime-limit": int(self.max_downtime * 1000),  # ms
                "max-bandwidth": self.migration_bandwidth,
            },
        )
        self.qmp.command("migrate", uri=address)
        self.log.debug(
            "migrate-parameters", **self.qmp.command("query-migrate-parameters")
        )

    def poll_migration_status(self, timeout=30):
        """Monitor ongoing migration.

        Every few seconds, the migration status is queried from the Qemu
        monitor. It is yielded to the calling context to provide a hook
        for communicating status updates.

        """
        timeout = TimeOut(timeout, 1, raise_on_timeout=True)
        while timeout.tick():
            if timeout.interval < 10:
                timeout.interval *= 1.4142
            info = self.qmp.command("query-migrate")
            yield info

            if info["status"] == "setup":
                pass
            elif info["status"] == "completed":
                break
            elif info["status"] == "active":
                # This didn't work out of the box on our 2.5, so I'll leave
                # this out for now. I think it's due to the need for the
                # userfaultd that needs to be installed on the host.
                # if info['ram']['transferred'] > info['ram']['total']:
                #     self.log.info('migrate-start-postcopy')
                #     self.qmp.command('migrate-start-postcopy')
                pass
            else:
                raise InvalidMigrationStatus(info)
            timeout.cutoff += 30

    def process_exists(self):
        proc = self.proc()
        if proc is None:
            return False
        return proc.is_running()

    def is_running(self):
        # This method must be very reliable. It is perfectly OK to error
        # out in the case of inconsistencies. But a "true" must mean:
        # we have a working Qemu instance here. And a "false" must mean:
        # there is no reason to think that any remainder of a Qemu process is
        # still running

        timeout = TimeOut(10, raise_on_timeout=False)
        while timeout.tick():
            # Try to find a stable result within a few seconds - ignore
            # unstable results in between. Qemu might just be starting
            # and the process already there but QMP not, or vice versa.

            # a) is there a process?
            expected_process_exists = self.process_exists()

            # b) is the monitor port around reliably? Let's assume it does.
            qmp_available = self.qmp

            # c) is the monitor available and talks to us?
            monitor_says_running = False
            status = {}

            if qmp_available:
                try:
                    status = self.qmp.command("query-status")
                except (QMPConnectError, socket.error):
                    # Force a reconnect in the next iteration.
                    self.__qmp.close()
                    self.__qmp = None
                    qmp_available = False
                    monitor_says_running = False
                else:
                    monitor_says_running = status["running"]

            if (
                expected_process_exists
                and qmp_available
                and monitor_says_running
            ):
                return True

            if not expected_process_exists and not qmp_available:
                return False

        # The timeout passed and we were not able to determine a consistent
        # result. :/
        raise VMStateInconsistent(
            "Can not determine whether Qemu is running. "
            "Process exists: {}, QMP socket reliable: {}, "
            "Status is running: {}".format(
                expected_process_exists, qmp_available, monitor_says_running
            ),
            status,
        )

    def rescue(self):
        """Recover from potentially inconsistent state.

        If the VM is running and we own all locks, then everything is fine.

        If the VM is running and we do not own the locks, then try to acquire
        them or bail out.

        Returns True if we were able to rescue the VM.
        Returns False if the rescue attempt failed and the VM is stopped now.

        """
        status = self.qmp.command("query-status")
        assert status["running"]
        for image in set(self.locks.available) - set(self.locks.held):
            try:
                self.acquire_lock(image)
            except Exception:
                self.log.debug("acquire-lock-failed", exc_info=True)
        self.assert_locks()

    def graceful_shutdown(self):
        if not self.qmp:
            return
        self.qmp.command(
            "send-key",
            keys=[
                {"type": "qcode", "data": "ctrl"},
                {"type": "qcode", "data": "alt"},
                {"type": "qcode", "data": "delete"},
            ],
        )

    def destroy(self):
        # We use this destroy command in "fire-and-forget"-style because
        # sometimes the init script will complain even if we achieve what
        # we want: that the VM isn't running any longer. We check this
        # by contacting the monitor instead.
        timeout = TimeOut(100, interval=1, raise_on_timeout=True)
        p = self.proc()
        if not p:
            return
        while p.is_running() and timeout.tick():
            try:
                p.terminate()
            except psutil.NoSuchProcess:
                pass

    def resize_root(self, size):
        self.qmp.command("block_resize", device="virtio0", size=size)

    def block_info(self):
        devices = {}
        for device in self.qmp.command("query-block"):
            devices[device["device"]] = device
        return devices

    def block_io_throttle(self, device, iops):
        self.qmp.command(
            "block_set_io_throttle",
            device=device,
            iops=iops,
            iops_rd=0,
            iops_wr=0,
            bps=0,
            bps_wr=0,
            bps_rd=0,
        )

    def watchdog_action(self, action):
        self.qmp.command(
            "human-monitor-command",
            **{"command-line": "watchdog_action action={}".format(action)},
        )

    def clean_run_files(self):
        runfiles = glob.glob("/run/qemu.{}.*".format(self.cfg["name"]))
        if not runfiles:
            return
        self.log.debug("clean-run-files")
        for runfile in runfiles:
            if runfile.endswith(".lock"):
                # Never, ever, remove lock files. Those should be on
                # partitions that get cleaned out during reboot, but
                # never otherwise.
                continue
            os.unlink(runfile)

    def prepare_config(self):
        chroot = self.chroot.format(**self.cfg)
        if not os.path.exists(chroot):
            os.makedirs(chroot)

        def format(s):
            return s.format(
                pidfile=self.pidfile,
                configfile=self.configfile,
                monitor_port=self.monitor_port,
                vnc=self.vnc.format(**self.cfg),
                chroot=chroot,
                **self.cfg,
            )

        self.local_args = [format(a) for a in self.args]
        self.local_config = format(self.config)

        with open(self.configfile + ".in", "w") as f:
            f.write(self.config)
        with open(self.configfile, "w") as f:
            f.write(self.local_config)
        with open(self.argfile + ".in", "w") as f:
            yaml.safe_dump(self.args, f)

        # Qemu tends to overwrite the pid file incompletely -> truncate
        open(self.pidfile, "w").close()

    def get_running_config(self):
        """Return the host-independent version of the current running
        config."""
        with open(self.argfile + ".in") as a:
            args = yaml.safe_load(a.read())
        with open(self.configfile + ".in") as c:
            config = c.read()
        return args, config

    def acquire_migration_lock(self):
        assert not self._migration_lock_fd
        open(self.migration_lockfile, "a+").close()
        self._migration_lock_fd = os.open(self.migration_lockfile, os.O_RDONLY)
        try:
            fcntl.flock(self._migration_lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            self.log.debug("acquire-migration-lock", result="success")
            return True
        except Exception as e:
            if isinstance(e, IOError):
                self.log.debug(
                    "acquire-migration-lock",
                    result="failure",
                    reason="competing lock",
                )
            else:
                self.log.exception(
                    "acquire-migration-lock", result="failure", exc_info=True
                )
            os.close(self._migration_lock_fd)
            self._migration_lock_fd = None
            return False

    def release_migration_lock(self):
        assert self._migration_lock_fd
        fcntl.flock(self._migration_lock_fd, fcntl.LOCK_UN)
        os.close(self._migration_lock_fd)
        self._migration_lock_fd = None



from setuptools import find_packages, setup

with open("README.txt") as f:
    readme = f.read() + "\n"
with open("CHANGES.txt") as f:
    readme += f.read() + "\n"
with open("HACKING.txt") as f:
    readme += f.read()

setup(
    name="fc.qemu",
    version="1.3.0",
    author="Christian Kauhaus, Christian Theune",
    author_email="mail@flyingcircus.io",
    url="http://github.com/flyingcircusio/fc.qemu",
    description="Qemu VM management utilities",
    long_description=readme,
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Programming Language :: Python :: 3.8",
    ],
    zip_safe=False,
    license="BSD",
    namespace_packages=["fc"],
    install_requires=[
        "colorama",  # ==0.3.3',
        "abaez.consulate==1.1.0",
        "psutil",  # ==5.4.2',
        "PyYaml>=5.3.1",
        "requests",  # ==2.11.1',
        "setuptools",
        "structlog>=16.1.0",
    ],
    entry_points={
        "console_scripts": [
            "fc-qemu = fc.qemu.main:main",
            "supervised-qemu = fc.qemu.hazmat.supervise:main",
        ],
    },
)



import pytest

from ..cpuscan import AbstractHost, AMDHost, IntelHost, scan_cpus


@pytest.mark.timeout(60)
@pytest.mark.slow
def test_cpuscan_intel():
    host = AMDHost()
    host.CPU_MODELS = ["qemu64-v1"]
    results = scan_cpus(host)
    for x in results:
        assert ",ssbd" not in x.cpu_arg
    assert "qemu64-v1" in [x.cpu_arg for x in results]


@pytest.mark.timeout(60)
@pytest.mark.slow
def test_cpuscan_amd():
    host = IntelHost()
    host.CPU_MODELS = ["qemu64-v1"]
    results = scan_cpus(host)
    for x in results:
        assert ",amd-ssbd" not in x.cpu_arg
    assert "qemu64-v1" in [x.cpu_arg for x in results]


@pytest.mark.timeout(60)
@pytest.mark.slow
def test_cpuscan_detects_something():
    host = AbstractHost()
    results = scan_cpus(host)
    assert [x.cpu_arg for x in results] == ["qemu64-v1"]



import os.path
import sys

from .util import ControlledRuntimeException, ensure_separate_cgroup


def daemonize(log=None):
    """
        Copyright/License abberation:

        Copied from
    http://stackoverflow.com/questions/1417631/python-code-to-daemonize-a-process
        and
    http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/

        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16

        With adjustments towards optional redirections of stdout and stderr
        to a specific file.
    """
    import os
    import os.path

    try:
        pid = os.fork()
        if pid > 0:
            # exit first parent
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

    # decouple from parent environment
    os.chdir("/")
    os.setsid()
    os.umask(0)

    # do second fork
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

    # redirect standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()
    si = open("/dev/null", "r")
    if log:
        so = se = log
    else:
        so = open("/dev/null", "a+")
        se = open("/dev/null", "a+")
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())


def main():
    ensure_separate_cgroup()
    import argparse

    a = argparse.ArgumentParser(description="Qemu VM agent")
    a.set_defaults(func="print_usage")

    a.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        default=False,
        help="Increase logging level.",
    )
    a.add_argument(
        "--daemonize",
        "-D",
        action="store_true",
        help="Run command in background.",
    )

    sub = a.add_subparsers(title="subcommands")

    p = sub.add_parser("ls", help="List VMs on this host.")
    p.set_defaults(func="ls")

    maint = sub.add_parser("maintenance", help="Perform maintenance tasks.")
    maint_sub = maint.add_subparsers()

    parser_enter = maint_sub.add_parser(
        "enter", help="Prepare host for maintenance mode."
    )
    parser_enter.set_defaults(func="maintenance_enter")

    p = sub.add_parser("check", help="Perform health check for this host.")
    p.set_defaults(func="check")

    p = sub.add_parser("status", help="Get the status of a VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="status")

    p = sub.add_parser("ensure", help="Ensure proper status of the VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="ensure")

    p = sub.add_parser("start", help="Start a VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="start")

    p = sub.add_parser("stop", help="Stop a VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="stop")

    p = sub.add_parser("restart", help="Restart a VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="restart")

    p = sub.add_parser("lock", help="Assume all locks of a VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="lock")

    p = sub.add_parser("unlock", help="Release all locks of a VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="unlock")

    p = sub.add_parser(
        "snapshot", help="Take a clean snapshot of this VM's root volume."
    )
    p.add_argument(
        "-k",
        "--keep",
        default=7,
        type=int,
        help="How many days to keep the snapshot. "
        "(default=7, forever=0) to keep",
    )
    p.add_argument("vm", metavar="VM", help="name of the VM")

    p.add_argument("snapshot", help="name of the snapshot")
    p.set_defaults(func="snapshot")

    p = sub.add_parser(
        "force-unlock",
        help="Release all locks of a VM even " "if we don't own them.",
    )
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="force_unlock")

    p = sub.add_parser("inmigrate", help="Start incoming migration for a VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="inmigrate")

    p = sub.add_parser("outmigrate", help="Start outgoing migration for a VM.")
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="outmigrate")

    p = sub.add_parser(
        "report-supported-cpu-models",
        help="Report the list of supported CPU models to the directory.",
    )
    p.set_defaults(func="report_supported_cpu_models")

    p = sub.add_parser(
        "shutdown-all", help="Cleanly shut down all VMS running on this host."
    )
    p.set_defaults(func="shutdown_all")

    p = sub.add_parser(
        "handle-consul-event",
        help="Handle a change in VM config distributed via consul.",
    )
    p.set_defaults(func="handle_consul_event")

    p = sub.add_parser(
        "telnet", help="Open a telnet connection to the VM's monitor port"
    )
    p.add_argument("vm", metavar="VM", help="name of the VM")
    p.set_defaults(func="telnet")

    args = a.parse_args()
    func = args.func

    if func == "print_usage":
        a.print_usage()
        sys.exit(1)

    vm = getattr(args, "vm", None)
    kwargs = dict(args._get_kwargs())
    del kwargs["func"]
    if "vm" in kwargs:
        del kwargs["vm"]
    del kwargs["daemonize"]
    del kwargs["verbose"]

    if args.daemonize:
        # Needed to help spawn subprocesses from consul without blocking.
        daemonize()

    from .agent import Agent, InvalidCommand, VMConfigNotFound
    from .logging import init_logging
    from .util import log

    try:
        init_logging(args.verbose)
        log.debug(" ".join(sys.argv))
        log.debug("load-system-config")

        from .sysconfig import sysconfig

        sysconfig.load_system_config()

        if vm is None:
            # Expecting a class/static method
            agent = Agent
            sys.exit(getattr(agent, func)(**kwargs) or 0)
        else:
            if "." in vm:
                # Our fc.agent calls us with path names. This is kinda stupid
                # but this was hidden in the Agent class before and now is
                # a bit more explicit.
                vm = os.path.basename(vm).split(".")[0]
            agent = Agent(vm)
            with agent:
                sys.exit(getattr(agent, func)(**kwargs) or 0)
    except (ControlledRuntimeException):
        # Those exceptions are properly logged and indicate an error with a
        # proper shutdown.
        log.exception("controlled-exception", exc_info=True)
        sys.exit(1)
    except (VMConfigNotFound, InvalidCommand):
        # Those exceptions are properly logged and don't have to be shown
        # with their traceback.
        log.debug("unexpected-exception")
        sys.exit(69)  # EX_UNAVAILABLE
    except Exception:
        log.exception("unexpected-exception", exc_info=True)
        sys.exit(69)  # EX_UNAVAILABLE



import os.path
import time

import pytest
import rbd

from ..volume import Volume


@pytest.fixture
def volume(ceph_inst):
    volume = Volume(ceph_inst, "othervolume", "label")
    volume.snapshots.purge()
    try:
        if volume._image:
            volume._image.close()
        rbd.RBD().remove(ceph_inst.ioctx, "othervolume")
    except rbd.ImageNotFound:
        pass

    yield volume
    time.sleep(0.2)

    lock = volume.lock_status()
    if lock is not None:
        volume.rbdimage.break_lock(*lock)
    volume.snapshots.purge()
    try:
        if volume._image:
            volume._image.close()
        rbd.RBD().remove(ceph_inst.ioctx, "othervolume")
    except rbd.ImageNotFound:
        pass


def test_volume_presence(volume):
    assert volume.fullname == "rbd.hdd/othervolume"
    assert not volume.exists()
    with pytest.raises(rbd.ImageNotFound):
        volume.rbdimage
    volume.ensure_presence()
    assert volume.rbdimage
    # Check that ensure_presence is fine with being called multiple times.
    volume.ensure_presence()


def test_volume_snapshot(volume):
    volume.ensure_presence()
    volume.snapshots.create("s0")
    snaps = list(volume.snapshots)
    assert len(snaps) == 1
    snapshot = snaps[0]
    assert snapshot.name == "othervolume"
    assert snapshot.snapname == "s0"
    assert snapshot.size == volume.size
    assert snapshot.id == volume.snapshots["s0"].id

    snapshot.remove()
    assert [] == list(volume.snapshots)


def test_purge_snapshots(volume):
    volume.ensure_presence()
    for snap in ["s0", "s1"]:
        volume.snapshots.create(snap)
    assert len(volume.snapshots) == 2
    volume.snapshots.purge()
    assert len(volume.snapshots) == 0


def test_snapshot_not_found(volume):
    with pytest.raises(KeyError):
        volume.snapshots["no-such-key"]


def test_volume_size(volume):
    volume.ensure_presence()
    assert volume.rbdimage
    assert volume.size == 1024
    volume.ensure_size(2048)
    assert volume.size == 2048
    # Call ensure multiple times to help triggering caching code paths.
    volume.ensure_size(2048)
    assert volume.size == 2048


def test_volume_shared_lock_protection(volume):
    volume.ensure_presence()
    volume.rbdimage.lock_shared("host1", "a")
    volume.rbdimage.lock_shared("remotehost", "a")
    with pytest.raises(NotImplementedError):
        volume.lock_status()
    lockers = volume.rbdimage.list_lockers()
    for client, cookie, _ in lockers["lockers"]:
        volume.rbdimage.break_lock(client, cookie)


def test_volume_locking(volume):
    # Non-existing volumes report None as locking status but do not raise
    # an exception.
    assert not volume.exists()
    assert volume.lock_status() is None
    volume.ensure_presence()
    assert volume.lock_status() is None
    volume.lock()
    assert volume.lock_status()[1] == "host1"
    # We want to smoothen out that some other process has locked the same image
    # for the same tag already and assume that this is another incarnation of
    # us - for that we have our own lock.
    volume.lock()
    assert volume.lock_status()[1] == "host1"
    volume.unlock()
    assert volume.lock_status() is None
    # We can call unlock twice if it isn't locked.
    volume.unlock()

    volume.rbdimage.lock_exclusive("someotherhost")
    with pytest.raises(rbd.ImageBusy):
        volume.lock()

    # Can not unlock locks that someone else holds.
    volume.unlock()
    assert volume.lock_status()[1] == "someotherhost"


def test_force_unlock(volume):
    volume.ensure_presence()
    volume.rbdimage.lock_exclusive("someotherhost")
    volume.unlock(force=True)
    assert volume.lock_status() is None


# increase timeout from the default of 3s
@pytest.mark.timeout(10)
def test_volume_mkswap(volume):
    volume.ensure_presence()
    volume.ensure_size(5 * 1024**2)
    with volume.mapped():
        volume.mkswap()


@pytest.mark.timeout(60)
def test_volume_mkfs(volume):
    volume.ensure_presence()
    volume.ensure_size(40 * 1024**2)
    with volume.mapped():
        volume.mkfs()


def test_unmapped_volume_should_have_no_part1(volume):
    # not something like 'None-part1'
    assert volume.part1dev is None


def test_volume_map_unmap_is_idempotent(volume):
    # This is more of an internal sanity test within our mocking infrastructure.
    volume.ensure_presence()
    volume.map()
    device = volume.device
    assert os.path.exists(device)
    volume.map()
    assert os.path.exists(device)
    volume.unmap()
    assert not os.path.exists(device)
    volume.unmap()
    assert not os.path.exists(device)


def test_map_snapshot(volume):
    volume.ensure_presence()
    volume.snapshots.create("s0")
    with volume.snapshots["s0"].mapped() as device:
        assert os.path.exists(device)


def test_mount_should_fail_if_not_mapped(volume):
    volume.ensure_presence()
    with pytest.raises(RuntimeError):
        volume.mount()


@pytest.mark.timeout(60)
@pytest.mark.live()
def test_mount_snapshot(volume):
    volume.ensure_presence()
    volume.ensure_size(40 * 1024**2)
    with volume.mapped():
        volume.mkfs(fstype="xfs", gptbios=True)
    volume.snapshots.create("s0")
    snap = volume.snapshots["s0"]
    with snap.mounted() as mp:
        mountpoint = mp
        assert os.path.ismount(mp)
        with open("/proc/self/mounts") as mounts:
            assert "{} xfs ro".format(mp) in mounts.read()
        # test for idempotence
        snap.mount()
        assert os.path.ismount(mp)
    assert not os.path.ismount(mountpoint)
    # test for idempotence
    snap.umount()
    assert not os.path.ismount(mountpoint)



import os
import shutil

import mock
import pkg_resources
import psutil
import pytest

from ..agent import Agent
from ..exc import VMStateInconsistent


@pytest.fixture
def cleanup_files():
    files = []
    yield files
    for f in files:
        if os.path.exists(f):
            os.unlink(f)


@pytest.fixture
def simplevm_cfg(cleanup_files):
    fixtures = pkg_resources.resource_filename(__name__, "fixtures")
    shutil.copy(fixtures + "/simplevm.yaml", "/etc/qemu/vm/simplevm.cfg")
    cleanup_files.append("/etc/qemu/vm/simplevm.cfg")
    yield "simplevm"


@pytest.mark.live
def test_builtin_config_template(simplevm_cfg):
    a = Agent(simplevm_cfg)
    with a:
        a.generate_config()
    # machine type must match Qemu version in virtualbox
    assert 'type = "pc-i440fx-4.1"' in a.qemu.config


@pytest.mark.live
def test_userdefined_config_template(simplevm_cfg, cleanup_files):
    with open("/etc/qemu/qemu.vm.cfg.in", "w") as f:
        f.write("# user defined config template\n")
    cleanup_files.append("/etc/qemu/qemu.vm.cfg.in")
    a = Agent(simplevm_cfg)
    with a:
        a.generate_config()
    assert "user defined config template" in a.qemu.config


@pytest.mark.live
@pytest.mark.timeout(60)
def test_consistency_vm_running(simplevm_cfg):
    a = Agent(simplevm_cfg)
    with a:
        a.qemu.is_running = mock.Mock(return_value=True)
        a.qemu.proc = mock.Mock(return_value=psutil.Process(1))
        a.ceph.locked_by_me = mock.Mock(return_value=True)
        a.raise_if_inconsistent()


@pytest.mark.live
@pytest.mark.timeout(60)
def test_consistency_vm_not_running(simplevm_cfg):
    a = Agent(simplevm_cfg)
    with a:
        a.qemu.is_running = mock.Mock(return_value=False)
        a.qemu.proc = mock.Mock(return_value=None)
        a.ceph.locked_by_me = mock.Mock(return_value=False)
        a.raise_if_inconsistent()


@pytest.mark.live
def test_consistency_process_dead(simplevm_cfg):
    a = Agent(simplevm_cfg)
    with a:
        a.qemu.is_running = mock.Mock(return_value=True)
        a.qemu.proc = mock.Mock(return_value=None)
        a.ceph.locked_by_me = mock.Mock(return_value=True)
        with pytest.raises(VMStateInconsistent):
            a.raise_if_inconsistent()


@pytest.mark.live
@pytest.mark.timeout(60)
def test_consistency_pidfile_missing(simplevm_cfg):
    a = Agent(simplevm_cfg)
    with a:
        a.qemu.is_running = mock.Mock(return_value=True)
        a.qemu.proc = mock.Mock(return_value=None)
        a.ceph.locked_by_me = mock.Mock(return_value=True)
        with pytest.raises(VMStateInconsistent):
            a.raise_if_inconsistent()


@pytest.mark.live
@pytest.mark.timeout(60)
def test_consistency_ceph_lock_missing(simplevm_cfg):
    a = Agent(simplevm_cfg)
    with a:
        a.qemu.is_running = mock.Mock(return_value=True)
        a.qemu.proc = mock.Mock(return_value=psutil.Process(1))
        a.ceph.locked_by_me = mock.Mock(return_value=False)
        with pytest.raises(VMStateInconsistent):
            a.raise_if_inconsistent()


@pytest.mark.live
@pytest.mark.timeout(60)
def test_ensure_inconsistent_state_detected(simplevm_cfg):
    a = Agent(simplevm_cfg)
    with a:
        a.qemu.is_running = mock.Mock(return_value=True)
        a.qemu.proc = mock.Mock(return_value=psutil.Process(1))
        a.ceph.locked_by_me = mock.Mock(return_value=False)
        with pytest.raises(VMStateInconsistent):
            a.raise_if_inconsistent()



# QEMU Monitor Protocol Python class
#
# Copyright (C) 2009, 2010 Red Hat Inc.
#
# Authors:
#  Luiz Capitulino <lcapitulino@redhat.com>
#
# This work is licensed under the terms of the GNU GPL, version 2.  See
# the COPYING file in the top-level directory.

import errno
import json
import socket
import sys


class QMPError(Exception):
    pass


class QMPConnectError(QMPError):
    pass


class QMPCapabilitiesError(QMPError):
    pass


class QMPTimeoutError(QMPError):
    pass


class QEMUMonitorProtocol:
    def __init__(self, address, log, server=False, debug=False):
        """
        Create a QEMUMonitorProtocol class.

        @param address: QEMU address, can be either a unix socket path (string)
                        or a tuple in the form ( address, port ) for a TCP
                        connection
        @param server: server mode listens on the socket (bool)
        @raise socket.error on socket connection errors
        @note No connection is established, this is done by the connect() or
              accept() methods
        """
        self.log = log.bind(subsystem="qemu/qmp")
        self.__events = []
        self.__address = address
        self._debug = debug
        self.__sock = self.__get_sock()
        if server:
            self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.__sock.bind(self.__address)
            self.__sock.listen(1)

    def __get_sock(self):
        if isinstance(self.__address, tuple):
            family = socket.AF_INET
        else:
            family = socket.AF_UNIX
        return socket.socket(family, socket.SOCK_STREAM)

    def __negotiate_capabilities(self):
        greeting = self.__json_read()
        if greeting is None or "QMP" not in greeting:
            raise QMPConnectError
        # Greeting seems ok, negotiate capabilities
        resp = self.cmd("qmp_capabilities")
        if "return" in resp:
            return greeting
        raise QMPCapabilitiesError

    def __json_read(self, only_event=False):
        while True:
            data = self.__sockfile.readline()
            if not data:
                return
            resp = json.loads(data)
            if "event" in resp:
                if self._debug:
                    print("QMP:<<< %s" % resp, file=sys.stderr)
                self.__events.append(resp)
                if not only_event:
                    continue
            return resp

    error = socket.error

    def __get_events(self, wait=False):
        """
        Check for new events in the stream and cache them in __events.

        @param wait (bool): block until an event is available.
        @param wait (float): If wait is a float, treat it as a timeout value.

        @raise QMPTimeoutError: If a timeout float is provided and the timeout
                                period elapses.
        @raise QMPConnectError: If wait is True but no events could be retrieved
                                or if some other error occurred.
        """

        # Check for new events regardless and pull them into the cache:
        self.__sock.setblocking(0)
        try:
            self.__json_read()
        except BlockingIOError:
            pass
        self.__sock.setblocking(1)

        # Wait for new events, if needed.
        # if wait is 0.0, this means "no wait" and is also implicitly false.
        if not self.__events and wait:
            if isinstance(wait, float):
                self.__sock.settimeout(wait)
            try:
                ret = self.__json_read(only_event=True)
            except socket.timeout:
                raise QMPTimeoutError("Timeout waiting for event")
            except Exception:
                raise QMPConnectError("Error while reading from socket")
            if ret is None:
                raise QMPConnectError("Error while reading from socket")
            self.__sock.settimeout(None)

    def connect(self, negotiate=True):
        """
        Connect to the QMP Monitor and perform capabilities negotiation.

        @return QMP greeting dict
        @raise socket.error on socket connection errors
        @raise QMPConnectError if the greeting is not received
        @raise QMPCapabilitiesError if fails to negotiate capabilities
        """
        self.__sock.connect(self.__address)
        self.__sockfile = self.__sock.makefile()
        if negotiate:
            return self.__negotiate_capabilities()

    def accept(self):
        """
        Await connection from QMP Monitor and perform capabilities negotiation.

        @return QMP greeting dict
        @raise socket.error on socket connection errors
        @raise QMPConnectError if the greeting is not received
        @raise QMPCapabilitiesError if fails to negotiate capabilities
        """
        self.__sock.settimeout(15)
        self.__sock, _ = self.__sock.accept()
        self.__sockfile = self.__sock.makefile()
        return self.__negotiate_capabilities()

    def cmd_obj(self, qmp_cmd):
        """
        Send a QMP command to the QMP Monitor.

        @param qmp_cmd: QMP command to be sent as a Python dict
        @return QMP response as a Python dict or None if the connection has
                been closed
        """
        self.log.debug(
            qmp_cmd["execute"],
            arguments=qmp_cmd.get("arguments", {}),
            id=qmp_cmd.get("id", None),
        )
        if self._debug:
            print("QMP:>>> %s" % qmp_cmd, file=sys.stderr)
        try:
            self.__sock.sendall(json.dumps(qmp_cmd).encode("ascii"))
        except BrokenPipeError:
            return
        resp = self.__json_read()
        if self._debug:
            print("QMP:<<< %s" % resp, file=sys.stderr)
        return resp

    def cmd(self, name, args=None, id=None):
        """
        Build a QMP command and send it to the QMP Monitor.

        @param name: command name (string)
        @param args: command arguments (dict)
        @param id: command id (dict, list, string or int)
        """
        qmp_cmd = {"execute": name}
        if args:
            qmp_cmd["arguments"] = args
        if id:
            qmp_cmd["id"] = id
        return self.cmd_obj(qmp_cmd)

    def command(self, cmd, **kwds):
        ret = self.cmd(cmd, kwds)
        if ret is None:
            raise QMPConnectError("Connection went away.")
        if "error" in ret:
            raise Exception(ret["error"]["desc"])
        return ret["return"]

    def pull_event(self, wait=False):
        """
        Get and delete the first available QMP event.

        @param wait (bool): block until an event is available.
        @param wait (float): If wait is a float, treat it as a timeout value.

        @raise QMPTimeoutError: If a timeout float is provided and the timeout
                                period elapses.
        @raise QMPConnectError: If wait is True but no events could be retrieved
                                or if some other error occurred.

        @return The first available QMP event, or None.
        """
        self.__get_events(wait)

        if self.__events:
            return self.__events.pop(0)
        return None

    def get_events(self, wait=False):
        """
        Get a list of available QMP events.

        @param wait (bool): block until an event is available.
        @param wait (float): If wait is a float, treat it as a timeout value.

        @raise QMPTimeoutError: If a timeout float is provided and the timeout
                                period elapses.
        @raise QMPConnectError: If wait is True but no events could be retrieved
                                or if some other error occurred.

        @return The list of available QMP events.
        """
        self.__get_events(wait)
        return self.__events

    def clear_events(self):
        """
        Clear current list of pending events.
        """
        self.__events = []

    def close(self):
        self.__sock.close()
        self.__sockfile.close()

    timeout = socket.timeout

    def settimeout(self, timeout):
        self.__sock.settimeout(timeout)

    def get_sock_fd(self):
        return self.__sock.fileno()

    def is_scm_available(self):
        return self.__sock.family == socket.AF_UNIX



from structlog import PrintLoggerFactory

from fc.qemu.logging import MultiOptimisticLoggerFactory, prefix


def test_prefix():
    assert (
        prefix("test00", "asdfasdfasdfa")
        == """\
test00>\tasdfasdfasdfa\
"""
    )
    assert (
        prefix("test00", "first line\nsecond line\nthird line\n")
        == """\
test00>\tfirst line
test00>\tsecond line
test00>\tthird line
test00>\t"""
    )


def test_multi_logger_ignores_error_on_fd(tmpdir):
    closed_stdout = open(str(tmpdir / "stdout"), "w")
    closed_stdout.close()
    f = MultiOptimisticLoggerFactory(console=PrintLoggerFactory(closed_stdout))
    logger = f()
    logger.msg(console="asdf")
